import logging
import re

from rpdk.core.exceptions import WizardAbortError, WizardValidationError
from rpdk.core.plugin_registry import get_plugin_choices
from rpdk.core.utils.init_utils import input_with_validation, print_error

LOG = logging.getLogger(__name__)
RESOURCE_TYPE_NAME_REGEX = r"^[a-zA-Z0-9]{2,64}::[a-zA-Z0-9]{2,64}::[a-zA-Z0-9]{2,64}$"


def init_resource(args, project):
    if args.type_name:
        try:
            type_name = validate_type_name(args.type_name)
        except WizardValidationError as error:
            print_error(error)
            type_name = input_typename()
    else:
        type_name = input_typename()

    if "language" in vars(args):
        language = args.language.lower()
    else:
        language = input_language()

    settings = {
        arg: getattr(args, arg)
        for arg in vars(args)
        if not callable(getattr(args, arg))
    }

    project.init(type_name, language, settings)
    project.generate()
    project.generate_docs()


def input_typename():
    type_name = input_with_validation(
        "What's the name of your resource type?",
        validate_type_name,
        "\n(Organization::Service::Resource)",
    )
    LOG.debug("Resource type identifier: %s", type_name)
    return type_name


def input_language():
    # language/plugin
    if validate_plugin_choice.max < 1:
        LOG.critical("No language plugins found")
        raise WizardAbortError()

    if validate_plugin_choice.max == 1:
        language = validate_plugin_choice.choices[0]
        LOG.warning("One language plugin found, defaulting to %s", language)
    else:
        language = input_with_validation(
            validate_plugin_choice.message, validate_plugin_choice
        )
    LOG.debug("Language plugin: %s", language)
    return language


def validate_type_name(value):
    match = re.match(RESOURCE_TYPE_NAME_REGEX, value)
    if match:
        return value
    LOG.debug("'%s' did not match '%s'", value, RESOURCE_TYPE_NAME_REGEX)
    raise WizardValidationError(
        "Please enter a value matching '{}'".format(RESOURCE_TYPE_NAME_REGEX)
    )


class ValidatePluginChoice:
    def __init__(self, choices):
        self.choices = tuple(choices)
        self.max = len(self.choices)

        pretty = "\n".join(
            "[{}] {}".format(i, choice) for i, choice in enumerate(self.choices, 1)
        )
        self.message = (
            "Select a language for code generation:\n"
            + pretty
            + "\n(enter an integer): "
        )

    def __call__(self, value):
        try:
            choice = int(value)
        except ValueError as e:
            raise WizardValidationError("Please enter an integer") from e
        choice -= 1
        if choice < 0 or choice >= self.max:
            raise WizardValidationError("Please select a choice")
        return self.choices[choice]


validate_plugin_choice = ValidatePluginChoice(  # pylint: disable=invalid-name
    get_plugin_choices()
)



FILTER_REGISTRY = {}


def register_filter(f):
    """Registers a filter function in this module's registry, so that filters
    can be added to Jinja2 environments easily.
    """
    FILTER_REGISTRY[f.__name__] = f
    return f


def parse_resource_type(resource_type):
    """Splits a resource type into it's components.

    :exc:`ValueError` is raised if the resource type is invalid.

    >>> parse_resource_type('AWS::ECS::Instance')
    ['AWS', 'ECS', 'Instance']
    >>> parse_resource_type('AWS::ECS')
    Traceback (most recent call last):
    ...
    ValueError: Resource type 'AWS::ECS' is invalid
    >>> parse_resource_type('AWS__ECS__Instance')
    Traceback (most recent call last):
    ...
    ValueError: Resource type 'AWS__ECS__Instance' is invalid
    """
    segments = resource_type.split("::")
    if len(segments) != 3:
        raise ValueError("Resource type '{}' is invalid".format(resource_type))
    return segments


@register_filter
def resource_type_namespace(resource_type):
    """Gets the namespace from a resource type.

    :exc:`ValueError` is raised if the resource type is invalid, see
    :func:`parse_resource_type`.

    >>> resource_type_namespace('AWS::ECS::Instance')
    'AWS'
    """
    return parse_resource_type(resource_type)[0]


@register_filter
def resource_type_service(resource_type):
    """Gets the service name from a resource type.

    :exc:`ValueError` is raised if the resource type is invalid, see
    :func:`parse_resource_type`.

    >>> resource_type_service('AWS::ECS::Instance')
    'ECS'
    """
    return parse_resource_type(resource_type)[1]


@register_filter
def resource_type_resource(resource_type):
    """Gets the resource name from a resource type.

    :exc:`ValueError` is raised if the resource type is invalid, see
    :func:`parse_resource_type`.

    >>> resource_type_resource('AWS::ECS::Instance')
    'Instance'
    """
    return parse_resource_type(resource_type)[2]


@register_filter
def lowercase_first_letter(string):
    """Converts the first letter of a string to lowercase.

    >>> lowercase_first_letter('CreateHandler')
    'createHandler'
    >>> lowercase_first_letter('createHandler')
    'createHandler'
    >>> lowercase_first_letter('')
    ''
    """
    return string[0].lower() + string[1:] if string else ""


@register_filter
def uppercase_first_letter(string):
    """Converts the first letter of a string to uppercase.

    >>> uppercase_first_letter('createHandler')
    'CreateHandler'
    >>> uppercase_first_letter('CreateHandler')
    'CreateHandler'
    >>> uppercase_first_letter('')
    ''
    """
    return string[0].upper() + string[1:] if string else ""


@register_filter
def modified_from_action_type(action_type):
    """Returns a boolean (as a string) representing whether the action type of
    the handler modifies the resource. Defaults to false.

    >>> modified_from_action_type('write')
    'true'
    >>> modified_from_action_type('read')
    'false'
    >>> modified_from_action_type('')
    'false'
    """
    if action_type == "write":
        return "true"
    return "false"


@register_filter
def package_prefix(full_package_name):
    """Returns the package prefix from the package name specified.

    :exc:`ValueError` is raised if the package name format is invalid.

    >>> package_prefix('com.example.test')
    'com.example'
    >>> package_prefix('example.test')
    'example'
    >>> package_prefix('com.example.this.isa.test')
    'com.example.this.isa'
    >>> package_prefix('exampletest')
    Traceback (most recent call last):
    ...
    ValueError: Package name 'exampletest' is invalid
    """
    package_segments = full_package_name.rpartition(".")
    if package_segments[0]:
        return package_segments[0]
    raise ValueError("Package name '{}' is invalid".format(full_package_name))



"""This sub command validates a project's resource schema.

Projects can be created via the 'init' sub command.
"""
import logging

from .project import Project

LOG = logging.getLogger(__name__)


def validate(_args):
    project = Project()
    project.load()


def setup_subparser(subparsers, parents):
    parser = subparsers.add_parser("validate", description=__doc__, parents=parents)
    parser.set_defaults(command=validate)



class RPDKBaseException(Exception):
    pass


class SysExitRecommendedError(RPDKBaseException):
    pass


class InternalError(RPDKBaseException):
    pass


class SpecValidationError(RPDKBaseException):
    pass


class FragmentValidationError(RPDKBaseException):
    pass


class WizardError(RPDKBaseException):
    pass


class WizardAbortError(WizardError):
    pass


class WizardValidationError(WizardError):
    pass


class UploadError(RPDKBaseException):
    pass


class InvalidProjectError(SysExitRecommendedError):
    pass


class CLIMisconfiguredError(SysExitRecommendedError):
    pass


class DownstreamError(RPDKBaseException):
    pass


class ContractTestError(RPDKBaseException):
    pass


class InvalidRequestError(ContractTestError):
    pass


class ModelResolverError(RPDKBaseException):
    pass


class InvalidFragmentFileError(RPDKBaseException):
    pass



# fixture and parameter have the same name
# pylint: disable=redefined-outer-name
import logging

import pytest

# WARNING: contract tests should use fully qualified imports to avoid issues
# when being loaded by pytest
from rpdk.core.contract.interface import Action, HandlerErrorCode, OperationStatus
from rpdk.core.contract.suite.handler_commons import (
    test_create_success,
    test_delete_failure_not_found,
    test_input_equals_output,
    test_model_in_list,
    test_read_failure_not_found,
    test_update_failure_not_found,
)

LOG = logging.getLogger(__name__)

# For the below tests, the deleted_model fixture
# is a resource that has been created and deleted


@pytest.fixture(scope="module")
def deleted_resource(resource_client):
    request = input_model = model = resource_client.generate_create_example()
    try:
        _status, response, _error = resource_client.call_and_assert(
            Action.CREATE, OperationStatus.SUCCESS, request
        )
        model = response["resourceModel"]
        test_input_equals_output(resource_client, input_model, model)
        _status, response, _error = resource_client.call_and_assert(
            Action.DELETE, OperationStatus.SUCCESS, model
        )
        assert (
            "resourceModel" not in response
        ), "The deletion handler's response object MUST NOT contain a model"
        yield model, request
    finally:
        status, response = resource_client.call(Action.DELETE, model)

        # a failed status is allowed if the error code is NotFound
        if status == OperationStatus.FAILED:
            error_code = resource_client.assert_failed(status, response)
            assert (
                error_code == HandlerErrorCode.NotFound
            ), "A delete hander MUST return FAILED with a NotFound error code\
                 if the resource did not exist prior to the delete request"
        else:
            resource_client.assert_success(status, response)


@pytest.mark.delete
@pytest.mark.read
def contract_delete_read(resource_client, deleted_resource):
    deleted_model, _request = deleted_resource
    test_read_failure_not_found(resource_client, deleted_model)


@pytest.mark.delete
@pytest.mark.list
def contract_delete_list(resource_client, deleted_resource):
    # LIST: Should not fail after deletion, since it is a list of all
    #       current resources of the resource type. Deletion should
    #       remove the model from the list, however.

    deleted_model, _request = deleted_resource
    assert not test_model_in_list(
        resource_client, deleted_model
    ), "A list operation MUST NOT return the primaryIdentifier \
        of any deleted resource instance"


@pytest.mark.delete
@pytest.mark.update
def contract_delete_update(resource_client, deleted_resource):
    deleted_model, _request = deleted_resource
    test_update_failure_not_found(resource_client, deleted_model)


@pytest.mark.delete
def contract_delete_delete(resource_client, deleted_resource):
    deleted_model, _request = deleted_resource
    test_delete_failure_not_found(resource_client, deleted_model)


@pytest.mark.create
@pytest.mark.delete
def contract_delete_create(resource_client, deleted_resource):
    if resource_client.has_only_writable_identifiers():
        _deleted_model, request = deleted_resource
        response = test_create_success(resource_client, request)

        resource_client.call_and_assert(
            Action.DELETE, OperationStatus.SUCCESS, response["resourceModel"]
        )
    else:
        pytest.skip("No writable identifiers. Skipping test.")



"""This sub command generates code from the project and resource schema.

Projects can be created via the 'init' sub command.
"""
import logging

from .project import Project

LOG = logging.getLogger(__name__)


def generate(_args):
    project = Project()
    project.load()
    project.generate()
    project.generate_docs()

    LOG.warning("Generated files for %s", project.type_name)


def setup_subparser(subparsers, parents):
    # see docstring of this file
    parser = subparsers.add_parser("generate", description=__doc__, parents=parents)
    parser.set_defaults(command=generate)



"""This sub command uploads the resource type to CloudFormation.

Projects can be created via the 'init' sub command.
"""
import logging

from .project import Project

LOG = logging.getLogger(__name__)


def submit(args):
    project = Project()
    project.load()
    project.submit(
        args.dry_run,
        args.endpoint_url,
        args.region,
        args.role_arn,
        args.use_role,
        args.set_default,
    )


def setup_subparser(subparsers, parents):
    # see docstring of this file
    parser = subparsers.add_parser("submit", description=__doc__, parents=parents)
    parser.set_defaults(command=submit)

    parser.add_argument(
        "--dry-run", action="store_true", help="Package the project, but do not submit."
    )
    parser.add_argument("--endpoint-url", help="CloudFormation endpoint to use.")
    parser.add_argument("--region", help="AWS Region to submit the resource type.")
    parser.add_argument(
        "--set-default",
        action="store_true",
        help="If registration is successful, set submitted version to the default.",
    )
    role_group = parser.add_mutually_exclusive_group()
    role_group.add_argument(
        "--role-arn",
        help="Role ARN that CloudFormation will use when invoking handlers.",
    )
    role_group.add_argument(
        "--no-role",
        action="store_false",
        dest="use_role",
        help="Register the type without an explicit execution role "
        "(Will not be able to invoke AWS APIs).",
    )



from itertools import count

from jsonschema.compat import urldefrag

from .pointer import fragment_decode


class RefRenamer:
    def __init__(self, renames=None):
        self.renames = renames if renames else {}
        # this generator never completes
        self.names = (
            name for name in ("schema{}".format(i) for i in count())
        )  # pragma: no cover

    def items(self):
        """Return all renames.

        >>> r = RefRenamer()
        >>> r.items()
        dict_items([])
        """
        return self.renames.items()

    def parse_ref_url(self, ref):
        """Parse a reference URL into a tuple of base + parts.

        Schemas are renamed for easier processing.

        >>> r = RefRenamer({"file://base.json": "base"})
        >>> sorted(r.items())
        [('file://base.json', 'base')]
        >>> r.parse_ref_url("file://base.json")
        ('base',)
        >>> r.parse_ref_url("file://base.json#/foo/bar")
        ('base', 'foo', 'bar')
        >>> sorted(r.items())
        [('file://base.json', 'base')]
        >>> r.parse_ref_url("file://other.json")
        ('schema0',)
        >>> r.parse_ref_url("file://other.json#/foo/bar")
        ('schema0', 'foo', 'bar')
        >>> sorted(r.items())
        [('file://base.json', 'base'), ('file://other.json', 'schema0')]
        """
        url, fragment = urldefrag(ref)
        try:
            rename = self.renames[url]
        except KeyError:
            self.renames[url] = rename = next(self.names)
        return (rename,) + fragment_decode(fragment, prefix="")



import pkg_resources

PLUGIN_REGISTRY = {
    entry_point.name: entry_point.load
    for entry_point in pkg_resources.iter_entry_points("rpdk.v1.languages")
}


def get_plugin_choices():
    plugin_choices = [
        entry_point.name
        for entry_point in pkg_resources.iter_entry_points("rpdk.v1.languages")
    ]
    return sorted(set(plugin_choices))


def get_parsers():
    parsers = {
        entry_point.name: entry_point.load
        for entry_point in pkg_resources.iter_entry_points("rpdk.v1.parsers")
    }

    return parsers


def load_plugin(language):
    return PLUGIN_REGISTRY[language]()()



import hashlib
import json
import logging
from collections.abc import Mapping, Sequence
from typing import Any, List, Tuple

from nested_lookup import nested_lookup
from ordered_set import OrderedSet

from .pointer import fragment_decode, fragment_encode

LOG = logging.getLogger(__name__)

NON_MERGABLE_KEYS = ("uniqueItems", "insertionOrder")
TYPE = "type"
REF = "$ref"
UNPACK_SEQUENCE_IDENTIFIER = "*"


class FlatteningError(Exception):
    pass


def item_hash(
    item,
):  # assumption -> input is only json comparable type (dict/list/scalar)
    """MD5 hash for an item (Dictionary/Iterable/Scalar)"""
    dhash = hashlib.md5()  # nosec
    if isinstance(item, dict):
        item = {k: item_hash(v) for k, v in item.items()}
    if isinstance(item, list):
        item = [item_hash(i) for i in item].sort()
    encoded = json.dumps(item, sort_keys=True).encode()
    dhash.update(encoded)
    return dhash.hexdigest()


def to_set(value: Any) -> OrderedSet:
    return (
        OrderedSet(value)
        if isinstance(value, (list, OrderedSet))
        else OrderedSet([value])
    )


class ConstraintError(FlatteningError, ValueError):
    def __init__(self, message, path, *args):
        self.path = fragment_encode(path)
        message = message.format(*args, path=self.path)
        super().__init__(message)


class BaseRefPlaceholder:
    """A sentinel object representing a reference inside the base document."""

    def __repr__(self):
        """Readable representation for debugging.

        >>> repr(BaseRefPlaceholder())
        '<BASE>'
        """
        return "<BASE>"


#: The sentinel instance representing a reference inside the base document.
BASE = BaseRefPlaceholder()


def rewrite_ref(ref):
    """Rewrite a reference to be inside of the base document. A relative JSON
    pointer is returned (in URI fragment identifier representation).

    If the reference is already inside the base document (:const:`BASE`), the parts
    are simply encoded into a pointer.

    If the reference is outside of the base document, a unique pointer inside
    the base document is made by namespacing the reference under the remote base
    name inside the remote section.

    >>> rewrite_ref((BASE, "foo", "bar"))
    '#/foo/bar'
    >>> rewrite_ref((BASE,))
    '#'
    >>> rewrite_ref(("remote", "foo", "bar"))
    '#/remote/remote/foo/bar'
    >>> rewrite_ref(("remote",))
    '#/remote/remote'
    """
    base, *parts = ref
    if base is not BASE:
        parts = ["remote", base] + parts
    return fragment_encode(parts)


def traverse(document, path_parts):
    """Traverse the document according to the reference.

    Since the document is presumed to be the reference's base, the base is
    discarded. There is no validation that the reference is valid.

    :raises ValueError, LookupError: the reference is invalid for this document

    >>> traverse({"foo": {"bar": [42]}}, tuple())
    ({'foo': {'bar': [42]}}, (), None)
    >>> traverse({"foo": {"bar": [42]}}, ["foo"])
    ({'bar': [42]}, ('foo',), {'foo': {'bar': [42]}})
    >>> traverse({"foo": {"bar": [42]}}, ("foo", "bar"))
    ([42], ('foo', 'bar'), {'bar': [42]})
    >>> traverse({"foo": {"bar": [42]}}, ("foo", "bar", "0"))
    (42, ('foo', 'bar', 0), [42])
    >>> traverse({}, ["foo"])
    Traceback (most recent call last):
    ...
    KeyError: 'foo'
    >>> traverse([], ["foo"])
    Traceback (most recent call last):
    ...
    ValueError: invalid literal for int() with base 10: 'foo'
    >>> traverse([], [0])
    Traceback (most recent call last):
    ...
    IndexError: list index out of range
    """
    parent = None
    path = []
    for part in path_parts:
        if isinstance(document, Sequence):
            part = int(part)
        parent = document
        document = document[part]
        path.append(part)
    return document, tuple(path), parent


def _resolve_ref(sub_schema, definitions):
    # resolve $ref
    ref = nested_lookup(REF, sub_schema)
    if ref:
        # [0] should be a single $ref in subschema on the top level
        # [-1] $ref must follow #/definitions/object
        sub_schema = definitions[fragment_decode(ref[0])[-1]]
    # resolve properties
    properties = nested_lookup("properties", sub_schema)
    if properties:
        sub_schema = properties[0]
    return sub_schema


# pylint: disable=C0301
def traverse_raw_schema(schema: dict, path: tuple):
    """Traverse the raw json schema resolving $ref

    :raises TypeError: either schema is not of type dict
    :raises ConstraintError: the schema tries to override "type" or "$ref"

    >>> traverse_raw_schema({"properties": {"bar": [42]}}, tuple())
    {'bar': [42]}
    >>> traverse_raw_schema({"properties": {"bar": [42]}}, ("bar",))
    [42]

    >>> traverse_raw_schema({"definitions": {"bar": {"type": "boolean"}},"properties": {"bar": {"$ref": "#/definitions/bar"}}}, ("bar",))
    {'type': 'boolean'}

    >>> traverse_raw_schema({"definitions":{"b":[1],"f":{"properties":{"b":{"$ref":"#/definitions/b"}}}},"properties":{"f":{"$ref":"#/definitions/f"}}},("f", "b")) # noqa: B950
    [1]

    >>> traverse_raw_schema({}, ("foo"))
    {}
    >>> traverse_raw_schema([], ["foo"])
    Traceback (most recent call last):
    ...
    TypeError: Schema must be a dictionary
    """
    if not isinstance(schema, Mapping):
        raise TypeError("Schema must be a dictionary")

    try:
        properties = schema["properties"]
        definitions = schema.get("definitions", {})
        sub_properties = properties
        for step in path:
            sub_properties = _resolve_ref(sub_properties[step], definitions)
        return sub_properties
    except KeyError as e:
        LOG.debug("Malformed Schema or incorrect path provided\n%s\n%s", path, e)
        return {}


def traverse_path_for_sequence_members(
    document: dict, path_parts: Sequence, path: list = None
) -> Tuple[List[object], List[tuple]]:
    """Traverse the paths for all sequence members in the document according to the reference.

    Since the document is presumed to be the reference's base, the base is
    discarded. There is no validation that the reference is valid.

    Differing from traverse, this returns a list of documents and a list of resolved paths.

    :parameter document: document to traverse (dict or list)
    :parameter path_parts: document paths to traverse
    :parameter path: traversed path so far

    :raises ValueError, LookupError: the reference is invalid for this document

    >>> traverse_path_for_sequence_members({"foo": {"bar": [42, 43, 44]}}, tuple())
    ([{'foo': {'bar': [42, 43, 44]}}], [()])
    >>> traverse_path_for_sequence_members({"foo": {"bar": [42, 43, 44]}}, ["foo"])
    ([{'bar': [42, 43, 44]}], [('foo',)])
    >>> traverse_path_for_sequence_members({"foo": {"bar": [42, 43, 44]}}, ("foo", "bar"))
    ([[42, 43, 44]], [('foo', 'bar')])
    >>> traverse_path_for_sequence_members({"foo": {"bar": [42, 43, 44]}}, ("foo", "bar", "*"))
    ([42, 43, 44], [('foo', 'bar', 0), ('foo', 'bar', 1), ('foo', 'bar', 2)])
    >>> traverse_path_for_sequence_members({"foo": {"bar": [{"baz": 1, "bin": 1}, {"baz": 2, "bin": 2}]}}, ("foo", "bar", "*"))
    ([{'baz': 1, 'bin': 1}, {'baz': 2, 'bin': 2}], [('foo', 'bar', 0), ('foo', 'bar', 1)])
    >>> traverse_path_for_sequence_members({"foo": {"bar": [{"baz": 1, "bin": 1}, {"baz": 2, "bin": 2}]}}, ("foo", "bar", "*", "baz"))
    ([1, 2], [('foo', 'bar', 0, 'baz'), ('foo', 'bar', 1, 'baz')])
    >>> traverse_path_for_sequence_members({}, ["foo"])
    Traceback (most recent call last):
    ...
    KeyError: 'foo'
    >>> traverse_path_for_sequence_members([], ["foo"])
    Traceback (most recent call last):
    ...
    ValueError: invalid literal for int() with base 10: 'foo'
    >>> traverse_path_for_sequence_members([], [0])
    Traceback (most recent call last):
    ...
    IndexError: list index out of range
    """
    if path is None:
        path = []
    if not path_parts:
        return [document], [tuple(path)]
    path_parts = list(path_parts)
    if not isinstance(document, Sequence):
        return _handle_non_sequence_for_traverse(document, path_parts, path)
    return _handle_sequence_for_traverse(document, path_parts, path)


def _handle_non_sequence_for_traverse(
    current_document: dict, current_path_parts: list, current_path: list
) -> Tuple[List[object], List[tuple]]:
    """
    Handling a non-sequence member for `traverse_path_for_sequence_members` is like the loop block in `traverse`:

    The next path part is the first part in the list of path parts.
    The new document is obtained from the current document using the new path part as the key.
    The next path part is added to the traversed path.

    The traversal continues by recursively calling `traverse_path_for_sequence_members`
    """
    part_to_handle = current_path_parts.pop(0)
    current_document = current_document[part_to_handle]
    current_path.append(part_to_handle)
    return traverse_path_for_sequence_members(
        current_document, current_path_parts, current_path
    )


def _handle_sequence_for_traverse(
    current_document: Sequence, current_path_parts: list, current_path: list
) -> Tuple[List[object], List[tuple]]:
    """
    Check the new path part for the unpack sequence identifier (e.g. '*'), otherwise traverse index and continue:

    The new document is obtained from the current document (a sequence) using the new path part as the index.
    The next path part is added to the traversed path
    """
    sequence_part = current_path_parts.pop(0)
    if sequence_part == UNPACK_SEQUENCE_IDENTIFIER:
        return _handle_unpack_sequence_for_traverse(
            current_document, current_path_parts, current_path
        )
    # otherwise, sequence part should be a valid index
    current_sequence_part = int(sequence_part)
    current_document = current_document[current_sequence_part]
    current_path.append(current_sequence_part)
    return [current_document], [tuple(current_path)]


def _handle_unpack_sequence_for_traverse(
    current_document: Sequence, current_path_parts: list, current_path: list
) -> Tuple[List[object], List[tuple]]:
    """
    When unpacking a sequence, we need to include multiple paths and multiple documents, one for each sequence member.

    For each sequence member:
    Append the traversed paths w/ the sequence index, and get the new document.
    The new document is obtained by traversing the current document using the sequence index.
    The new document is appended to the list of new documents.

    For each new document:
    The remaining document is traversed using the remaining path parts.
    The list of traversed documents and traversed paths are returned.
    """
    documents = []
    resolved_paths = []
    new_documents = []
    new_paths = []
    for sequence_index in range(len(current_document)):
        new_paths.append(current_path.copy() + [sequence_index])
        new_document = traverse_path_for_sequence_members(
            current_document, [sequence_index] + current_path_parts, current_path.copy()
        )[0]
        new_documents.extend(new_document)
    for i in range(len(new_documents)):  # pylint: disable=consider-using-enumerate
        new_document = new_documents[i]
        newer_documents, newer_paths = traverse_path_for_sequence_members(
            new_document, current_path_parts, new_paths[i]
        )
        documents.extend(newer_documents)
        resolved_paths.extend(newer_paths)
    return documents, resolved_paths


def schema_merge(target, src, path):  # noqa: C901 # pylint: disable=R0912
    """Merges the src schema into the target schema in place.

    If there are duplicate keys, src will overwrite target.

    :raises TypeError: either schema is not of type dict
    :raises ConstraintError: the schema tries to override "type" or "$ref"

    >>> schema_merge({}, {}, ())
    {}
    >>> schema_merge({'foo': 'a'}, {}, ())
    {'foo': 'a'}

    >>> schema_merge({}, {'foo': 'a'}, ())
    {'foo': 'a'}

    >>> schema_merge({'foo': 'a'}, {'foo': 'b'}, ())
    {'foo': 'b'}

    >>> schema_merge({'required': 'a'}, {'required': 'b'}, ())
    {'required': ['a', 'b']}

    >>> a, b = {'$ref': 'a'}, {'foo': 'b'}
    >>> schema_merge(a, b, ('foo',))
    {'$ref': 'a', 'foo': 'b'}

    >>> a, b = {'$ref': 'a'}, {'type': 'b'}
    >>> schema_merge(a, b, ('foo',))
    {'type': OrderedSet(['a', 'b'])}

    >>> a, b = {'$ref': 'a'}, {'$ref': 'b'}
    >>> schema_merge(a, b, ('foo',))
    {'type': OrderedSet(['a', 'b'])}

    >>> a, b = {'$ref': 'a'}, {'type': ['b', 'c']}
    >>> schema_merge(a, b, ('foo',))
    {'type': OrderedSet(['a', 'b', 'c'])}

    >>> a, b = {'$ref': 'a'}, {'type': OrderedSet(['b', 'c'])}
    >>> schema_merge(a, b, ('foo',))
    {'type': OrderedSet(['a', 'b', 'c'])}

    >>> a, b = {'type': ['a', 'b']}, {'$ref': 'c'}
    >>> schema_merge(a, b, ('foo',))
    {'type': OrderedSet(['a', 'b', 'c'])}

    >>> a, b = {'type': OrderedSet(['a', 'b'])}, {'$ref': 'c'}
    >>> schema_merge(a, b, ('foo',))
    {'type': OrderedSet(['a', 'b', 'c'])}

    >>> a, b = {'Foo': {'$ref': 'a'}}, {'Foo': {'type': 'b'}}
    >>> schema_merge(a, b, ('foo',))
    {'Foo': {'type': OrderedSet(['a', 'b'])}}

    >>> schema_merge({'type': 'a'}, {'type': 'b'}, ()) # doctest: +NORMALIZE_WHITESPACE
    {'type': OrderedSet(['a', 'b'])}

    >>> schema_merge({'type': 'string'}, {'type': 'integer'}, ())
    {'type': OrderedSet(['string', 'integer'])}
    """
    if not (isinstance(target, Mapping) and isinstance(src, Mapping)):
        raise TypeError("Both schemas must be dictionaries")

    for key, src_schema in src.items():
        try:
            if key in (
                REF,
                TYPE,
            ):  # $ref and type are treated similarly and unified
                target_schema = target.get(key) or target.get(TYPE) or target[REF]
            else:
                target_schema = target[key]  # carry over existing properties
        except KeyError:
            target[key] = src_schema
        else:
            next_path = path + (key,)
            try:
                target[key] = schema_merge(target_schema, src_schema, next_path)
            except TypeError:
                if key in (TYPE, REF):  # combining multiple $ref and types
                    src_set = to_set(src_schema)

                    try:
                        target[TYPE] = to_set(
                            target[TYPE]
                        )  # casting to ordered set as lib
                        # implicitly converts strings to sets
                        target[TYPE] |= src_set
                    except (TypeError, KeyError):
                        target_set = to_set(target_schema)
                        target[TYPE] = target_set | src_set

                    try:
                        # check if there are conflicting $ref and type
                        # at the same sub schema. Conflicting $ref could only
                        # happen on combiners because method merges two json
                        # objects without losing any previous info:
                        # e.g. "oneOf": [{"$ref": "..#1.."},{"$ref": "..#2.."}] ->
                        # { "ref": "..#1..", "type": [{},{}] }
                        target.pop(REF)
                    except KeyError:
                        pass

                elif key == "required":
                    target[key] = sorted(set(target_schema) | set(src_schema))
                else:
                    if key in NON_MERGABLE_KEYS and target_schema != src_schema:
                        msg = (
                            "Object at path '{path}' declared multiple values "
                            "for '{}': found '{}' and '{}'"
                        )
                        # pylint: disable=W0707
                        raise ConstraintError(msg, path, key, target_schema, src_schema)
                    target[key] = src_schema
    return target



from functools import wraps
from inspect import Parameter, signature

import pytest

from rpdk.core.contract.interface import HandlerErrorCode


def _rebind(decorator, func, *args, **kwargs):
    """Helper function to construct decorated arguments

    This works only with positional and likely positional arguments
    strongly keyword arguments are in **kwargs. It constructs kwargs'
    from positional values
    """
    parameters = signature(func).parameters.values()
    decorated_parameters = set(signature(decorator).parameters.keys())

    positional_kwargs = dict(
        zip(
            [
                parameter.name
                for parameter in parameters
                if parameter.kind
                in (Parameter.POSITIONAL_ONLY, Parameter.POSITIONAL_OR_KEYWORD)
                and parameter.name not in kwargs
            ],
            args,
        )
    )
    return {k: kwargs.get(k) or positional_kwargs[k] for k in decorated_parameters}


def decorate(after=True):
    """Helper function to construct decorator from a simple function

    arg: after means that decorated check should be run after the
    target function

    This is a 'decorate' meta function that wraps new decorator around
    target function and merges decorated arguments with target arguments
    convention: each new decorator should have a 'response' argument,
    which is an output of a target function
    """

    def inner_decorator(decorator: object):
        def new_decorator(func: object):
            @wraps(func)
            def function(*args, **kwargs):
                response_arg = {}
                if after:  # running function before the decorated check
                    response = func(*args, **kwargs)  # calling target function
                    response_arg = {"response": response}

                kvargs = _rebind(decorator, func, *args, **{**kwargs, **response_arg})
                decorated_sig = signature(decorator)
                bound_arguments = decorated_sig.bind(**kvargs)
                decorator(
                    *bound_arguments.args, **bound_arguments.kwargs
                )  # calling a decorated function to execute check

                # this allows to make a pre-execution check
                # e.g. if skip function
                if not after:  # running function after the decorated check
                    response = func(*args, **kwargs)  # calling target function
                return response

            return function

        return new_decorator

    return inner_decorator


@decorate()
def response_does_not_contain_write_only_properties(resource_client, response):
    resource_client.assert_write_only_property_does_not_exist(response["resourceModel"])


@decorate()
def response_contains_resource_model_equal_updated_model(
    response, current_resource_model, update_resource_model
):
    assert response["resourceModel"] == {
        **current_resource_model,
        **update_resource_model,
    }, "All properties specified in the update request MUST be present in the \
        model returned, and they MUST match exactly, with the exception of \
            properties defined as writeOnlyProperties in the resource schema"


@decorate()
def response_contains_primary_identifier(resource_client, response):
    resource_client.assert_primary_identifier(
        resource_client.primary_identifier_paths, response["resourceModel"]
    )


@decorate()
def response_contains_unchanged_primary_identifier(
    resource_client, response, current_resource_model
):
    assert resource_client.is_primary_identifier_equal(
        resource_client.primary_identifier_paths,
        current_resource_model,
        response["resourceModel"],
    ), "PrimaryIdentifier returned in every progress event must match \
        the primaryIdentifier passed into the request"


@decorate(after=False)
def skip_not_writable_identifier(resource_client):
    if not resource_client.has_only_writable_identifiers():
        pytest.skip("No writable identifiers. Skipping test.")


@decorate(after=False)
def skip_no_tagging(resource_client):
    if not resource_client.contains_tagging_metadata():
        pytest.skip("Resource does not contain tagging metadata. Skipping test.")


@decorate(after=False)
def skip_not_taggable(resource_client):
    if not resource_client.is_taggable():
        pytest.skip("Resource is not taggable. Skipping test.")


@decorate(after=False)
def skip_not_tag_updatable(resource_client):
    if not resource_client.is_tag_updatable():
        pytest.skip("Resource is not tagUpdatable. Skipping test.")


def failed_event(error_code, msg=""):
    def decorator_wrapper(func: object):
        @wraps(func)
        def wrapper(*args, **kwargs):
            response_error = func(*args, **kwargs)
            if response_error is not None:
                if isinstance(error_code, HandlerErrorCode):
                    error_code_tuple = (error_code,)
                assert response_error in error_code_tuple, msg
            return response_error

        return wrapper

    return decorator_wrapper



import logging
from collections.abc import Iterable, Mapping

from jsonschema import RefResolver

from .renamer import RefRenamer
from .utils import BASE, rewrite_ref, traverse

LOG = logging.getLogger(__name__)


class RefInliner(RefResolver):
    """Mutates the schema."""

    def __init__(self, base_uri, schema):
        self.schema = schema
        self.ref_graph = {}

        # our meta-schema should catch this, but better to be explicit
        if "remote" in self.schema:
            raise ValueError("Schema already contains remote schemas.")

        self.renamer = RefRenamer(renames={base_uri: BASE})
        super().__init__(base_uri=base_uri, referrer=self.schema, cache_remote=True)

    def _walk_schema(self):
        self._walk(self.schema, (BASE,))

    def _walk(self, obj, old_path):
        if isinstance(obj, str):
            return  # very common, easier to debug this case

        if isinstance(obj, Mapping):
            for key, value in obj.items():
                if key == "$ref":
                    if old_path in self.ref_graph:
                        LOG.debug("Already visited '%s' (%s)", old_path, value)
                        return
                    url, resolved = self.resolve(value)
                    LOG.debug("Resolved '%s' to '%s'", value, url)
                    # parse the URL into
                    new_path = self.renamer.parse_ref_url(url)
                    LOG.debug("Parsed '%s' to '%s'", url, new_path)
                    LOG.debug("Edge from '%s' to '%s'", old_path, new_path)
                    self.ref_graph[old_path] = new_path
                    self.push_scope(url)
                    try:
                        self._walk(resolved, new_path)
                    finally:
                        self.pop_scope()
                else:
                    self._walk(value, old_path + (key,))
        # order matters, both Mapping and strings are also Iterable
        elif isinstance(obj, Iterable):
            for i, value in enumerate(obj):
                self._walk(value, old_path + (str(i),))
        # fall-through: for other types, there's nothing to do

    def _rewrite_refs(self):
        for base_uri, rename in self.renamer.items():
            LOG.debug("Rewriting refs in '%s' (%s)", rename, base_uri)
            document = self.store[base_uri]
            for from_ref, to_ref in self.ref_graph.items():
                base, *parts = from_ref
                # only process refs in this file
                if base != rename:
                    continue
                current, _path, _parent = traverse(document, parts)
                new_ref = rewrite_ref(to_ref)
                LOG.debug("  '%s' -> '%s'", current["$ref"], new_ref)
                current["$ref"] = new_ref

    def _inline_defs(self):
        global_defs = {}
        for base_uri, rename in self.renamer.items():
            if rename is BASE:  # no need to process the local file
                continue
            LOG.debug("Inlining definitions from '%s' (%s)", rename, base_uri)
            global_defs[rename] = local_defs = {"$comment": base_uri}
            local_defs.update(self.store[base_uri])
        if global_defs:
            self.schema["remote"] = global_defs

    def inline(self):
        self._walk_schema()
        self._rewrite_refs()
        self._inline_defs()
        return self.schema



# pylint: disable=import-outside-toplevel
# pylint: disable=R0904
import json
import logging
import re
import time
from time import sleep
from uuid import uuid4

import docker
from botocore import UNSIGNED
from botocore.config import Config

from rpdk.core.contract.interface import Action, HandlerErrorCode, OperationStatus
from rpdk.core.contract.type_configuration import TypeConfiguration
from rpdk.core.exceptions import InvalidProjectError

from ..boto_helpers import (
    LOWER_CAMEL_CRED_KEYS,
    create_sdk_session,
    get_account,
    get_temporary_credentials,
)
from ..jsonutils.pointer import fragment_decode, fragment_list
from ..jsonutils.utils import (
    UNPACK_SEQUENCE_IDENTIFIER,
    item_hash,
    traverse,
    traverse_path_for_sequence_members,
    traverse_raw_schema,
)

LOG = logging.getLogger(__name__)
LOOKUP_ERROR_MESSAGE_FORMAT = (
    "Caught LookupError when pruning properties for document %s and path %s"
)


def prune_properties(document, paths):
    """Prune given properties from a document.

    This assumes properties will always have an object (dict) as a parent.
    The function modifies the document in-place, but also returns the document
    for convenience. (The return value may be ignored.)
    """
    for path in paths:
        # if '*' is in path, we need to prune more than one property (prune property for all members of array)
        if UNPACK_SEQUENCE_IDENTIFIER in path:
            document = _prune_properties_for_all_sequence_members(document, path)
            continue
        try:
            _prop, resolved_path, parent = traverse(document, path)
        except LookupError:
            # not found means nothing to delete
            LOG.info(LOOKUP_ERROR_MESSAGE_FORMAT, document, path)
        else:
            key = resolved_path[-1]
            del parent[key]
    return document


def _prune_properties_for_all_sequence_members(document: dict, path: list) -> dict:
    try:
        # this returns multiple paths
        _prop, resolved_paths = traverse_path_for_sequence_members(document, path)
    except LookupError:
        # not found means nothing to delete
        LOG.info(LOOKUP_ERROR_MESSAGE_FORMAT, document, path)
    else:
        # paths with indices are gathered in increasing order, but we need to prune in reverse order
        resolved_paths = resolved_paths[::-1]
        for resolved_path in resolved_paths:
            new_doc = document
            for key in resolved_path[: len(resolved_path) - 1]:
                new_doc = new_doc[key]
            key = resolved_path[-1]
            del new_doc[key]
    return document


def prune_properties_if_not_exist_in_path(output_model, input_model, paths):
    """Prune given properties from a model.

    This assumes properties will always have an object (dict) as a parent.
    The function returns the model after pruning the path which exists
    in the paths tuple but not in the input_model
    """
    output_document = {"properties": output_model.copy()}
    input_document = {"properties": input_model.copy()}
    for path in paths:
        try:
            if not path_exists(input_document, path):
                _prop, resolved_path, parent = traverse(output_document, path)
                key = resolved_path[-1]
                del parent[key]
        except LookupError:
            pass
    return output_document["properties"]


def prune_properties_which_dont_exist_in_path(model, paths):
    """Prunes model to properties present in path. This method removes any property
    from the model which does not exists in the paths

    This assumes properties will always have an object (dict) as a parent.
    The function returns the model after pruning all but the path which exists
    in the paths tuple from the input_model
    """
    document = {"properties": model.copy()}
    for model_path in model.keys():
        path_tuple = ("properties", model_path)
        if path_tuple not in paths:
            _prop, resolved_path, parent = traverse(document, path_tuple)
            key = resolved_path[-1]
            del parent[key]
    return document["properties"]


def path_exists(document, path):
    try:
        _prop, _resolved_path, _parent = traverse(document, path)
    except LookupError:
        return False
    else:
        return True


def prune_properties_from_model(model, paths):
    """Prune given properties from a resource model.

    This assumes the dict passed in is a resources model i.e. solely the properties.
    This also assumes the paths to remove are prefixed with "/properties",
    as many of the paths are defined in the schema
    The function modifies the document in-place, but also returns the document
    for convenience. (The return value may be ignored.)
    """
    return prune_properties({"properties": model}, paths)["properties"]


def override_properties(document, overrides):
    """Override given properties from a document."""
    for path, obj in overrides.items():
        try:
            _prop, resolved_path, parent = traverse(document, path)
        except LookupError:
            LOG.debug(
                "Override failed.\nPath %s\nObject %s\nDocument %s", path, obj, document
            )
            LOG.warning("Override with path %s not found, skipping", path)
        else:
            key = resolved_path[-1]
            parent[key] = obj
    return document


class ResourceClient:  # pylint: disable=too-many-instance-attributes
    def __init__(
        self,
        function_name,
        endpoint,
        region,
        schema,
        overrides,
        inputs=None,
        role_arn=None,
        timeout_in_seconds="30",
        type_name=None,
        log_group_name=None,
        log_role_arn=None,
        docker_image=None,
        executable_entrypoint=None,
    ):  # pylint: disable=too-many-arguments
        self._session = create_sdk_session(region)
        self._role_arn = role_arn
        self._type_name = type_name
        self._log_group_name = log_group_name
        self._log_role_arn = log_role_arn
        self.region = region
        self.account = get_account(
            self._session,
            get_temporary_credentials(self._session, LOWER_CAMEL_CRED_KEYS, role_arn),
        )
        self._function_name = function_name
        if endpoint.startswith("http://"):
            self._client = self._session.client(
                "lambda",
                endpoint_url=endpoint,
                use_ssl=False,
                verify=False,
                config=Config(
                    signature_version=UNSIGNED,
                    # needs to be long if docker is running on a slow machine
                    read_timeout=15 * 60,
                    retries={"max_attempts": 0},
                    region_name=self._session.region_name,
                ),
            )
        else:
            self._client = self._session.client("lambda", endpoint_url=endpoint)

        self._schema = None
        self._strategy = None
        self._update_strategy = None
        self._invalid_strategy = None
        self._overrides = overrides
        self._update_schema(schema)
        self._inputs = inputs
        self._timeout_in_seconds = int(timeout_in_seconds)
        self._docker_image = docker_image
        self._docker_client = docker.from_env() if self._docker_image else None
        self._executable_entrypoint = executable_entrypoint

    def _properties_to_paths(self, key):
        return {fragment_decode(prop, prefix="") for prop in self._schema.get(key, [])}

    def _update_schema(self, schema):
        # TODO: resolve $ref
        self._schema = schema
        self._strategy = None
        self._update_strategy = None
        self._invalid_strategy = None

        self.primary_identifier_paths = self._properties_to_paths("primaryIdentifier")
        self.read_only_paths = self._properties_to_paths("readOnlyProperties")
        self.write_only_paths = self._properties_to_paths("writeOnlyProperties")
        self.create_only_paths = self._properties_to_paths("createOnlyProperties")
        self.properties_without_insertion_order = self.get_metadata()

        additional_identifiers = self._schema.get("additionalIdentifiers", [])
        self._additional_identifiers_paths = [
            {fragment_decode(prop, prefix="") for prop in identifier}
            for identifier in additional_identifiers
        ]

    def has_only_writable_identifiers(self):
        return all(
            path in self.create_only_paths for path in self.primary_identifier_paths
        )

    def assert_write_only_property_does_not_exist(self, resource_model):
        if self.write_only_paths:
            assert not any(
                self.key_error_safe_traverse(resource_model, write_only_property)
                for write_only_property in self.write_only_paths
            ), "The model MUST NOT return properties defined as \
                writeOnlyProperties in the resource schema"

    def get_metadata(self):
        try:
            properties = self._schema["properties"]
        except KeyError:
            return set()
        else:
            return {
                prop
                for prop in properties.keys()
                if "insertionOrder" in properties[prop]
                and properties[prop]["insertionOrder"] == "false"
            }

    @property
    def strategy(self):
        # an empty strategy (i.e. false-y) is valid
        if self._strategy is not None:
            return self._strategy

        # imported here to avoid hypothesis being loaded before pytest is loaded
        from .resource_generator import ResourceGenerator

        # make a copy so the original schema is never modified
        schema = json.loads(json.dumps(self._schema))

        prune_properties(schema, self.read_only_paths)

        self._strategy = ResourceGenerator(schema).generate_schema_strategy(schema)
        return self._strategy

    @property
    def invalid_strategy(self):
        # an empty strategy (i.e. false-y) is valid
        if self._invalid_strategy is not None:
            return self._invalid_strategy

        # imported here to avoid hypothesis being loaded before pytest is loaded
        from .resource_generator import ResourceGenerator

        # make a copy so the original schema is never modified
        schema = json.loads(json.dumps(self._schema))

        self._invalid_strategy = ResourceGenerator(schema).generate_schema_strategy(
            schema
        )
        return self._invalid_strategy

    @property
    def update_strategy(self):
        # an empty strategy (i.e. false-y) is valid
        if self._update_strategy is not None:
            return self._update_strategy

        # imported here to avoid hypothesis being loaded before pytest is loaded
        from .resource_generator import ResourceGenerator

        # make a copy so the original schema is never modified
        schema = json.loads(json.dumps(self._schema))

        prune_properties(schema, self.read_only_paths)
        prune_properties(schema, self.create_only_paths)

        self._update_strategy = ResourceGenerator(schema).generate_schema_strategy(
            schema
        )
        return self._update_strategy

    def generate_create_example(self):
        if self._inputs:
            return self._inputs["CREATE"]
        example = self.strategy.example()
        return override_properties(example, self._overrides.get("CREATE", {}))

    def generate_invalid_create_example(self):
        if self._inputs:
            return self._inputs["INVALID"]
        example = self.invalid_strategy.example()
        return override_properties(example, self._overrides.get("CREATE", {}))

    def get_unique_keys_for_model(self, create_model):
        return {
            k: v
            for k, v in create_model.items()
            if self.is_property_in_path(k, self.primary_identifier_paths)
            or any(
                map(
                    lambda additional_identifier_paths, key=k: self.is_property_in_path(
                        key, additional_identifier_paths
                    ),
                    self._additional_identifiers_paths,
                )
            )
        }

    @staticmethod
    def is_property_in_path(key, paths):
        for path in paths:
            prop = fragment_list(path, "properties")[0]
            if prop == key:
                return True
        return False

    @staticmethod
    def get_value_by_key_path(model, key_path):
        if isinstance(key_path, (tuple, list)):
            value = model
            for path in key_path:
                value = value[path]
            return value
        return model[key_path]

    def validate_update_example_keys(self, unique_identifiers, update_example):
        for primary_identifier in self.primary_identifier_paths:
            if primary_identifier in self.create_only_paths:
                primary_key_path = fragment_list(primary_identifier, "properties")
                update_example_pk_value = self.get_value_by_key_path(
                    update_example, primary_key_path
                )
                unique_identifiers_pk_value = self.get_value_by_key_path(
                    unique_identifiers, primary_key_path
                )
                assert update_example_pk_value == unique_identifiers_pk_value, (
                    "Any createOnlyProperties specified in update handler input "
                    "MUST NOT be different from their previous state"
                )

    def generate_update_example(self, create_model):
        if self._inputs:
            unique_identifiers = self.get_unique_keys_for_model(create_model)
            update_example = self._inputs["UPDATE"]
            if self.create_only_paths:
                self.validate_update_example_keys(unique_identifiers, update_example)
            update_example.update(unique_identifiers)
            create_model_with_read_only_properties = (
                prune_properties_which_dont_exist_in_path(
                    create_model, self.read_only_paths
                )
            )
            return {**create_model_with_read_only_properties, **update_example}
        overrides = self._overrides.get("UPDATE", self._overrides.get("CREATE", {}))
        example = override_properties(self.update_strategy.example(), overrides)
        return {**create_model, **example}

    def generate_invalid_update_example(self, create_model):
        if self._inputs:
            return self._inputs["INVALID"]
        overrides = self._overrides.get("UPDATE", self._overrides.get("CREATE", {}))
        example = override_properties(self.invalid_strategy.example(), overrides)
        return {**create_model, **example}

    def compare(self, inputs, outputs, path=()):
        assertion_error_message = (
            "All properties specified in the request MUST "
            "be present in the model returned, and they MUST"
            " match exactly, with the exception of properties"
            " defined as writeOnlyProperties in the resource schema"
        )
        try:
            if isinstance(inputs, dict):
                for key in inputs:
                    new_path = path + (key,)
                    if isinstance(inputs[key], dict):
                        self.compare(inputs[key], outputs[key], new_path)
                    elif isinstance(inputs[key], list):
                        assert len(inputs[key]) == len(outputs[key])

                        is_ordered = traverse_raw_schema(self._schema, new_path).get(
                            "insertionOrder", True
                        )
                        self.compare_collection(
                            inputs[key], outputs[key], is_ordered, new_path
                        )
                    else:
                        assert inputs[key] == outputs[key], assertion_error_message
            else:
                assert inputs == outputs, assertion_error_message
        except Exception as exception:
            raise AssertionError(assertion_error_message) from exception

    def compare_collection(self, inputs, outputs, is_ordered, path):
        if is_ordered:
            for index in range(len(inputs)):  # pylint: disable=C0200
                self.compare(inputs[index], outputs[index], path)
            return

        assert {item_hash(item) for item in inputs} == {
            item_hash(item) for item in outputs
        }

    @staticmethod
    def key_error_safe_traverse(resource_model, write_only_property):
        try:
            return traverse(
                resource_model, fragment_list(write_only_property, "properties")
            )[0]
        except KeyError:
            return None

    @staticmethod
    def assert_in_progress(status, response):
        assert status == OperationStatus.IN_PROGRESS, "status should be IN_PROGRESS"
        assert (
            response.get("errorCode", 0) == 0
        ), "IN_PROGRESS events should have no error code set"
        assert (
            response.get("resourceModels") is None
        ), "IN_PROGRESS events should not include any resource models"

        return response.get("callbackDelaySeconds", 0)

    @staticmethod
    def assert_success(status, response):
        assert status == OperationStatus.SUCCESS, "status should be SUCCESS"
        assert (
            response.get("errorCode", 0) == 0
        ), "SUCCESS events should have no error code set"
        assert (
            response.get("callbackDelaySeconds", 0) == 0
        ), "SUCCESS events should have no callback delay"

    @staticmethod
    def assert_failed(status, response):
        assert status == OperationStatus.FAILED, "status should be FAILED"
        assert "errorCode" in response, "FAILED events must have an error code set"
        # raises a KeyError if the error code is invalid
        error_code = HandlerErrorCode[response["errorCode"]]
        assert (
            response.get("callbackDelaySeconds", 0) == 0
        ), "FAILED events should have no callback delay"
        assert (
            response.get("resourceModels") is None
        ), "FAILED events should not include any resource models"

        return error_code

    @staticmethod
    # pylint: disable=R0913
    def make_request(
        desired_resource_state,
        previous_resource_state,
        region,
        account,
        action,
        creds,
        type_name,
        log_group_name,
        log_creds,
        token,
        callback_context=None,
        type_configuration=None,
        **kwargs,
    ):
        request_body = {
            "requestData": {
                "callerCredentials": creds,
                "resourceProperties": desired_resource_state,
                "previousResourceProperties": previous_resource_state,
                "logicalResourceId": token,
                "typeConfiguration": type_configuration,
            },
            "region": region,
            "awsAccountId": account,
            "action": action,
            "callbackContext": callback_context,
            "bearerToken": token,
            "resourceType": type_name,
            **kwargs,
        }
        if log_group_name and log_creds:
            request_body["requestData"]["providerCredentials"] = log_creds
            request_body["requestData"]["providerLogGroupName"] = log_group_name
        return request_body

    @staticmethod
    def generate_token():
        return str(uuid4())

    def assert_time(self, start_time, end_time, action):
        timeout_in_seconds = (
            self._timeout_in_seconds
            if action in (Action.READ, Action.LIST)
            else self._timeout_in_seconds * 2
        )
        assert end_time - start_time <= timeout_in_seconds, (
            "Handler %r timed out." % action
        )

    @staticmethod
    def assert_primary_identifier(primary_identifier_paths, resource_model):
        try:
            assert all(
                traverse(
                    resource_model, fragment_list(primary_identifier, "properties")
                )[0]
                for primary_identifier in primary_identifier_paths
            ), "Every returned model MUST include the primaryIdentifier"
        except KeyError as e:
            raise AssertionError(
                "Every returned model MUST include the primaryIdentifier"
            ) from e

    @staticmethod
    def is_primary_identifier_equal(
        primary_identifier_path, created_model, updated_model
    ):
        try:
            return all(
                traverse(
                    created_model, fragment_list(primary_identifier, "properties")
                )[0]
                == traverse(
                    updated_model, fragment_list(primary_identifier, "properties")
                )[0]
                for primary_identifier in primary_identifier_path
            )
        except KeyError as e:
            raise AssertionError(
                "The primaryIdentifier returned in every progress event must\
                     match the primaryIdentifier passed into the request"
            ) from e

    def _make_payload(
        self,
        action,
        current_model,
        previous_model=None,
        type_configuration=None,
        **kwargs,
    ):
        return self.make_request(
            current_model,
            previous_model,
            self.region,
            self.account,
            action,
            get_temporary_credentials(
                self._session, LOWER_CAMEL_CRED_KEYS, self._role_arn
            ),
            self._type_name,
            self._log_group_name,
            get_temporary_credentials(
                self._session, LOWER_CAMEL_CRED_KEYS, self._log_role_arn
            ),
            self.generate_token(),
            type_configuration=type_configuration,
            **kwargs,
        )

    def _call(self, payload):
        request_without_write_properties = prune_properties(
            payload["requestData"]["resourceProperties"], self.write_only_paths
        )

        previous_request_without_write_properties = None
        if payload["requestData"]["previousResourceProperties"]:
            previous_request_without_write_properties = prune_properties(
                payload["requestData"]["previousResourceProperties"],
                self.write_only_paths,
            )
        payload_to_log = {
            "callbackContext": payload["callbackContext"],
            "action": payload["action"],
            "requestData": {
                "resourceProperties": request_without_write_properties,
                "previousResourceProperties": previous_request_without_write_properties,
                "logicalResourceId": payload["requestData"]["logicalResourceId"],
            },
            "region": payload["region"],
            "awsAccountId": payload["awsAccountId"],
            "bearerToken": payload["bearerToken"],
        }
        LOG.debug(
            "Sending request\n%s",
            json.dumps(payload_to_log, ensure_ascii=False, indent=2),
        )
        payload = json.dumps(payload, ensure_ascii=False, indent=2)
        if self._docker_image:
            if not self._executable_entrypoint:
                raise InvalidProjectError(
                    "executableEntrypoint not set in .rpdk-config. "
                    "Have you run cfn generate?"
                )
            result = (
                self._docker_client.containers.run(
                    self._docker_image,
                    self._executable_entrypoint + " '" + payload + "'",
                )
                .decode()
                .strip()
            )
            LOG.debug("=== Handler execution logs ===")
            LOG.debug(result)
            # pylint: disable=W1401
            regex = "__CFN_RESOURCE_START_RESPONSE__([\s\S]*)__CFN_RESOURCE_END_RESPONSE__"  # noqa: W605,B950 # pylint: disable=C0301
            payload = json.loads(re.search(regex, result).group(1))
        else:
            result = self._client.invoke(
                FunctionName=self._function_name, Payload=payload.encode("utf-8")
            )

            try:
                payload = json.load(result["Payload"])
            except json.decoder.JSONDecodeError as json_error:
                LOG.debug("Received invalid response\n%s", result["Payload"])
                raise ValueError(
                    "Handler Output is not a valid JSON document"
                ) from json_error
        LOG.debug("Received response\n%s", payload)
        return payload

    def call_and_assert(
        self, action, assert_status, current_model, previous_model=None, **kwargs
    ):
        if not self.has_required_handlers():
            raise ValueError("Create/Read/Delete handlers are required")
        if assert_status not in [OperationStatus.SUCCESS, OperationStatus.FAILED]:
            raise ValueError("Assert status {} not supported.".format(assert_status))

        status, response = self.call(action, current_model, previous_model, **kwargs)
        if assert_status == OperationStatus.SUCCESS:
            self.assert_success(status, response)
            error_code = None
        else:
            error_code = self.assert_failed(status, response)
        return status, response, error_code

    def call(self, action, current_model, previous_model=None, **kwargs):
        request = self._make_payload(
            action,
            current_model,
            previous_model,
            TypeConfiguration.get_type_configuration(),
            **kwargs,
        )
        start_time = time.time()
        response = self._call(request)
        self.assert_time(start_time, time.time(), action)

        # this throws a KeyError if status isn't present, or if it isn't a valid status
        status = OperationStatus[response["status"]]

        if action in (Action.READ, Action.LIST):
            assert status != OperationStatus.IN_PROGRESS
            return status, response

        while status == OperationStatus.IN_PROGRESS:
            callback_delay_seconds = self.assert_in_progress(status, response)
            self.assert_primary_identifier(
                self.primary_identifier_paths, response.get("resourceModel")
            )
            sleep(callback_delay_seconds)

            request["requestData"]["resourceProperties"] = response.get("resourceModel")
            request["callbackContext"] = response.get("callbackContext")
            # refresh credential for every handler invocation
            request["requestData"]["callerCredentials"] = get_temporary_credentials(
                self._session, LOWER_CAMEL_CRED_KEYS, self._role_arn
            )

            response = self._call(request)
            status = OperationStatus[response["status"]]

        # ensure writeOnlyProperties are not returned on final responses
        if "resourceModel" in response.keys() and status == OperationStatus.SUCCESS:
            self.assert_write_only_property_does_not_exist(response["resourceModel"])

        return status, response

    def has_update_handler(self):
        return "update" in self._schema["handlers"]

    def contains_tagging_metadata(self):
        return "tagging" in self._schema

    def is_taggable(self):
        try:
            return self._schema["tagging"]["taggable"]
        except KeyError:
            try:
                return self._schema["taggable"]
            except KeyError:
                return True

    def is_tag_updatable(self):
        try:
            return self._schema["tagging"]["tagUpdatable"]
        except KeyError:
            return self.is_taggable()

    def metadata_contains_tag_property(self):
        try:
            return "tagProperty" in self._schema["tagging"]
        except KeyError:
            return False

    def validate_model_contain_tags(self, inputs):
        assertion_error_message = "Contract test inputs does not contain tags property."
        try:
            tag_property_path = self._schema["tagging"]["tagProperty"]
        except KeyError:
            tag_property_path = "/properties/Tags"
        tag_property_name = tag_property_path.split("/")[-1]
        LOG.debug("Defined tag property name is: %s\n", tag_property_name)
        try:
            if isinstance(inputs, dict):
                for key in inputs:
                    if key == tag_property_name:
                        return True
            else:
                raise assertion_error_message
        except Exception as exception:
            raise AssertionError(assertion_error_message) from exception
        return False

    def has_required_handlers(self):
        try:
            has_delete = "delete" in self._schema["handlers"]
            has_create = "create" in self._schema["handlers"]
            has_read = "read" in self._schema["handlers"]
            return has_read and has_create and has_delete
        except KeyError:
            return False



import json
import logging
import os

from rpdk.core.exceptions import InvalidProjectError

LOG = logging.getLogger(__name__)

TYPE_CONFIGURATION_FILE_PATH = "~/.cfn-cli/typeConfiguration.json"


class TypeConfiguration:
    TYPE_CONFIGURATION = None

    @staticmethod
    def get_type_configuration():
        LOG.debug(
            "Loading type configuration setting file at '~/.cfn-cli/typeConfiguration.json'"
        )
        if TypeConfiguration.TYPE_CONFIGURATION is None:
            try:
                with open(
                    os.path.expanduser(TYPE_CONFIGURATION_FILE_PATH), encoding="utf-8"
                ) as f:
                    TypeConfiguration.TYPE_CONFIGURATION = json.load(f)
            except json.JSONDecodeError as json_decode_error:
                LOG.debug(
                    "Type configuration file '%s' is invalid",
                    TYPE_CONFIGURATION_FILE_PATH,
                )
                raise InvalidProjectError(
                    "Type configuration file '%s' is invalid"
                    % TYPE_CONFIGURATION_FILE_PATH
                ) from json_decode_error
            except FileNotFoundError:
                LOG.debug(
                    "Type configuration file '%s' not Found, do nothing",
                    TYPE_CONFIGURATION_FILE_PATH,
                )
        return TypeConfiguration.TYPE_CONFIGURATION



"""
This class has two responsibilities:
1. generating a sample template fragment so the user has some initial
file fragment as an example.
The method "generate_sample_fragment" will be called as part of the init command.
2. generating schema for provided template fragments.
The method "generate_schema" will be called right before submission.
"""
import json
import logging
import os
from pathlib import Path

from rpdk.core.data_loaders import resource_json
from rpdk.core.exceptions import FragmentValidationError

from .lint_warning_printer import print_cfn_lint_warnings
from .module_fragment_reader import get_template_file_size_in_bytes, read_raw_fragments

LOG = logging.getLogger(__name__)
FRAGMENT_DIR = "fragments"
SAMPLE_FRAGMENT_OUTPUT = "sample.json"
SCHEMA_NAME = "schema.json"
SAMPLE_FRAGMENT = "../data/examples/module/sample.json"
RESOURCE_LIMIT = 500
OUTPUT_LIMIT = 200
MAPPING_LIMIT = 200
MAPPING_ATTRIBUTE_LIMIT = 200
TEMPLATE_FILE_SIZE_IN_BYTES_LIMIT = 1500000


class TemplateFragment:  # pylint: disable=too-many-instance-attributes
    def __init__(self, type_name, root=None):
        self.root = Path(root) if root else Path.cwd()
        self.fragment_dir = self.root / FRAGMENT_DIR
        self.type_name = type_name
        self.resource_limit = RESOURCE_LIMIT
        self.output_limit = OUTPUT_LIMIT
        self.mapping_limit = MAPPING_LIMIT
        self.mapping_attribute_limit = MAPPING_ATTRIBUTE_LIMIT
        self.template_file_size_in_bytes_limit = TEMPLATE_FILE_SIZE_IN_BYTES_LIMIT

        LOG.debug("Fragment directory: %s", self.fragment_dir)

    def generate_schema(self):
        raw_fragments = read_raw_fragments(self.fragment_dir)

        schema = {}
        properties = {}

        schema["typeName"] = self.type_name
        schema["description"] = "Schema for Module Fragment of type " + self.type_name
        schema["properties"] = properties
        schema["additionalProperties"] = True

        if "Parameters" in raw_fragments:
            properties["Parameters"] = self.__build_parameters(raw_fragments)
        properties["Resources"] = self.__build_resources(raw_fragments)

        self.__write_schema(schema)

        return schema

    def validate_fragments(self):
        """
        This method makes sure that the fragments adhere
            to the template fragment restrictions.
        Note: Fn::ImportValue was checked when loading the fragments
            since it can occur anywhere in the template.
        """
        raw_fragments = read_raw_fragments(self.fragment_dir)
        self.__validate_file_size_limit()
        self.__validate_resources(raw_fragments)
        self.__validate_parameters(raw_fragments)
        self.__validate_no_transforms_present(raw_fragments)
        self.__validate_outputs(raw_fragments)
        self.__validate_mappings(raw_fragments)
        print_cfn_lint_warnings(self.fragment_dir)

    def __validate_outputs(self, raw_fragments):
        self.__validate_no_exports_present(raw_fragments)
        self.__validate_output_limit(raw_fragments)

    @staticmethod
    def __validate_no_exports_present(raw_fragments):
        if "Outputs" in raw_fragments:
            for _output_name, output in raw_fragments["Outputs"].items():
                if "Export" in output:
                    raise FragmentValidationError(
                        "Template fragment cannot contain any Export. "
                        "Found an Export statement in Output: " + _output_name
                    )

    def __validate_output_limit(self, raw_fragments):
        if "Outputs" in raw_fragments:
            output_count = len(raw_fragments["Outputs"].items())
            if output_count > self.output_limit:
                raise FragmentValidationError(
                    "The Module template fragment has "
                    + str(output_count)
                    + " outputs but must not exceed the limit of "
                    + str(self.output_limit)
                    + " outputs"
                )

    def __validate_resources(self, raw_fragments):
        if "Resources" not in raw_fragments:
            raise FragmentValidationError(
                "A Module template fragment must have a Resources section"
            )
        self.__validate_resource_limit(raw_fragments)
        for _resource_name, resource in raw_fragments["Resources"].items():
            if "Type" in resource:
                self.__validate_no_nested_stacks(resource)
                self.__validate_no_macros(resource)
            elif "Name" in resource:
                self.__validate_no_include(resource)
                raise FragmentValidationError(
                    "Resource '" + _resource_name + "' is invalid"
                )
            else:
                raise FragmentValidationError(
                    "Resource '" + _resource_name + "' has neither Type nor Name"
                )

    @staticmethod
    def __validate_no_include(resource):
        if resource["Name"] == "AWS::Include":
            raise FragmentValidationError(
                "Template fragment can't use AWS::Include transform."
            )

    @staticmethod
    def __validate_no_macros(resource):
        if resource["Type"] == "AWS::CloudFormation::Macro":
            raise FragmentValidationError("Template fragment can't contain any macro.")

    @staticmethod
    def __validate_no_nested_stacks(resource):
        if resource["Type"] == "AWS::CloudFormation::Stack":
            raise FragmentValidationError(
                "Template fragment can't contain nested stack."
            )

    def __validate_resource_limit(self, raw_fragments):
        resource_count = len(raw_fragments["Resources"].items())
        if resource_count > self.resource_limit:
            raise FragmentValidationError(
                "The Module template fragment has "
                + str(resource_count)
                + " resources but must not exceed the limit of "
                + str(self.resource_limit)
                + " resources"
            )

    @staticmethod
    def __validate_parameters(raw_fragments):
        if "Parameters" in raw_fragments:
            for _parameter_name, parameter in raw_fragments["Parameters"].items():
                if "Type" not in parameter:
                    raise FragmentValidationError(
                        "Parameter '" + _parameter_name + "' must have a Type"
                    )

    @staticmethod
    def __validate_no_transforms_present(raw_fragments):
        if "transform" in raw_fragments or "Transform" in raw_fragments:
            raise FragmentValidationError(
                "Template fragment can't contain transform section."
            )
        if "Fn::Transform" in raw_fragments:
            raise FragmentValidationError(
                "Template fragment can't contain any transform."
            )

    def __validate_mappings(self, raw_fragments):
        self.__validate_mapping_limit(raw_fragments)
        self.__validate_mapping_attribute_limit(raw_fragments)

    def __validate_mapping_limit(self, raw_fragments):
        if "Mappings" in raw_fragments:
            mapping_count = len(raw_fragments["Mappings"].items())
            if mapping_count > self.mapping_limit:
                raise FragmentValidationError(
                    "The Module template fragment has "
                    + str(mapping_count)
                    + " mappings but must not exceed the limit of "
                    + str(self.output_limit)
                    + " mappings"
                )

    def __validate_mapping_attribute_limit(self, raw_fragments):
        if "Mappings" in raw_fragments:
            for _mapping_name, mapping in raw_fragments["Mappings"].items():
                attribute_count = len(mapping.items())
                if attribute_count > self.mapping_attribute_limit:
                    raise FragmentValidationError(
                        "The mapping "
                        + _mapping_name
                        + " has "
                        + str(attribute_count)
                        + " attributes but must not exceed the limit of "
                        + str(self.output_limit)
                        + " mapping attributes"
                    )

    def __validate_file_size_limit(self):
        total_size = get_template_file_size_in_bytes(self.fragment_dir)
        if total_size > self.template_file_size_in_bytes_limit:
            raise FragmentValidationError(
                "The total file size of the template"
                " fragments exceeds the CloudFormation Template size limit"
            )

    @staticmethod
    def __build_resources(raw_fragments):
        raw_resources = {}
        resources = {}
        for resource in raw_fragments["Resources"]:
            raw_resources[resource] = {
                "type": raw_fragments["Resources"][resource]["Type"]
            }
        resources_properties = {}
        for resource in raw_resources:
            type_object = {"type": "object", "properties": {}}
            type_object["properties"]["Type"] = {
                "type": "string",
                "const": raw_resources[resource]["type"],
            }
            type_object["properties"]["Properties"] = {"type": "object"}
            resources_properties[resource] = type_object
        resources["properties"] = resources_properties
        resources["type"] = "object"
        resources["additionalProperties"] = False
        return resources

    @staticmethod
    def __build_parameters(raw_fragments):
        raw_parameters = {}
        parameters = {}
        for param in raw_fragments["Parameters"]:
            param_type = raw_fragments["Parameters"][param]["Type"]

            description = raw_fragments["Parameters"][param].get("Description")
            raw_parameters[param] = {
                "type": param_type.lower(),
                "description": description,
            }
        parameter_properties = {}
        for raw_param in raw_parameters:
            description = raw_parameters[raw_param]["description"]
            type_name = "object"
            properties = {"Type": {"type": "string"}}
            required = ["Type"]
            parameter_properties[raw_param] = {
                "type": type_name,
                "properties": properties,
                "required": required,
            }
            if description is not None:
                parameter_properties[raw_param]["description"] = description
                properties["Description"] = {"type": "string"}
                required.append("Description")
        parameters["type"] = "object"
        parameters["properties"] = parameter_properties
        return parameters

    def __write_schema(self, schema):
        def _write(f):
            json.dump(schema, f, indent=4)
            f.write("\n")

        self._overwrite(self.root / SCHEMA_NAME, _write)

    def generate_sample_fragment(self):
        self._create_fragment_directory()
        sample_json = self.__get_sample_fragment_json()

        def _write(f):
            json.dump(sample_json, f, indent=4)
            f.write("\n")

        self._overwrite(self.fragment_dir / SAMPLE_FRAGMENT_OUTPUT, _write)

    @staticmethod
    def __get_sample_fragment_json():
        sample_json = resource_json(__name__, SAMPLE_FRAGMENT)
        return sample_json

    def _create_fragment_directory(self):
        if not os.path.exists(self.fragment_dir):
            os.mkdir(self.fragment_dir)
            print("Directory ", self.fragment_dir, " Created ")
        else:
            print("Directory ", self.fragment_dir, " already exists")

    @staticmethod
    def _overwrite(path, contents):
        LOG.debug("Overwriting '%s'", path)
        with path.open("w", encoding="utf-8") as f:
            if callable(contents):
                contents(f)
            else:
                f.write(contents)



import pytest


class ContractPlugin:
    def __init__(self, resource_client):
        self._resource_client = resource_client

    @pytest.fixture(scope="module")
    def resource_client(self):
        return self._resource_client



# fixture and parameter have the same name
# pylint: disable=redefined-outer-name

import pytest

# WARNING: contract tests should use fully qualified imports to avoid issues
# when being loaded by pytest
from rpdk.core.contract.interface import Action, HandlerErrorCode, OperationStatus
from rpdk.core.contract.suite.contract_asserts import failed_event


@pytest.mark.update
@failed_event(
    error_code=HandlerErrorCode.NotFound,
    msg="An update handler MUST return FAILED with a NotFound error code\
         if the resource did not exist prior to the update request",
)
def contract_update_without_create(resource_client):
    create_request = resource_client.generate_create_example()
    update_request = resource_client.generate_update_example(create_request)
    _status, response, _error = resource_client.call_and_assert(
        Action.UPDATE, OperationStatus.FAILED, update_request, create_request
    )
    assert response[
        "message"
    ], "The progress event MUST return an error message\
         when the status is failed"
    return _error



import logging
import re

from rpdk.core.exceptions import WizardValidationError
from rpdk.core.fragment.generator import TemplateFragment
from rpdk.core.utils.init_utils import input_with_validation, print_error

LOG = logging.getLogger(__name__)

# this regex has to be kept in sync with the one in the meta-schema.
MODULE_TYPE_NAME_REGEX = (
    r"^[a-zA-Z0-9]{2,64}::[a-zA-Z0-9]{2,64}::[a-zA-Z0-9]{2,64}::MODULE$"
)


def init_module(args, project):
    if args.type_name:
        try:
            type_name = validate_type_name(args.type_name)
        except WizardValidationError as error:
            print_error(error)
            type_name = input_typename()
    else:
        type_name = input_typename()

    project.init_module(type_name)
    template_fragment = TemplateFragment(type_name, project.root)
    template_fragment.generate_sample_fragment()


def input_typename():
    type_name = input_with_validation(
        "What's the name of your module type?",
        validate_type_name,
        "\n(<Organization>::<Service>::<Name>::MODULE)",
    )
    LOG.debug("Resource type identifier: %s", type_name)
    return type_name


def validate_type_name(value):
    match = re.match(MODULE_TYPE_NAME_REGEX, value)
    if match:
        return value
    LOG.debug("'%s' did not match '%s'", value, MODULE_TYPE_NAME_REGEX)
    raise WizardValidationError(
        "Please enter a value matching '{}'".format(MODULE_TYPE_NAME_REGEX)
    )



import pytest


def contract_check_asserts_work():
    message = (
        "Asserts have been stripped. This is unusual, but happens if the "
        "contract tests are compiled to optimized byte code. As a result, the "
        "contract tests will not run correctly. Please raise an issue with "
        "as much information about your system and runtime as possible."
    )
    with pytest.raises(AssertionError):
        assert False  # noqa: B011
        pytest.fail(message)



import logging
from datetime import datetime

from botocore.exceptions import ClientError, WaiterError

from .data_loaders import resource_stream
from .exceptions import DownstreamError, InternalError, InvalidProjectError, UploadError

LOG = logging.getLogger(__name__)

BUCKET_OUTPUT_NAME = "CloudFormationManagedUploadBucketName"
LOG_DELIVERY_ROLE_ARN_OUTPUT_NAME = "LogAndMetricsDeliveryRoleArn"
EXECUTION_ROLE_ARN_OUTPUT_NAME = "ExecutionRoleArn"
INFRA_STACK_NAME = "CloudFormationManagedUploadInfrastructure"


class Uploader:
    def __init__(self, cfn_client, s3_client):
        self.cfn_client = cfn_client
        self.s3_client = s3_client
        self.bucket_name = ""
        self.log_delivery_role_arn = ""

    @staticmethod
    def _get_template():
        with resource_stream(__name__, "data/managed-upload-infrastructure.yaml") as f:
            template = f.read()

        # sanity test! it's super easy to rename one but not the other
        for output_name in [BUCKET_OUTPUT_NAME, LOG_DELIVERY_ROLE_ARN_OUTPUT_NAME]:
            if output_name not in template:
                LOG.debug(
                    "Output '%s' not found in managed upload "
                    "infrastructure template:\n%s",
                    output_name,
                    template,
                )
                raise InternalError(
                    "Output not found in managed upload infrastructure template"
                )

        return template

    def _wait_for_stack(self, stack_id, waiter_name, success_msg):
        waiter = self.cfn_client.get_waiter(waiter_name)
        LOG.debug("Waiting for stack '%s'", stack_id)
        try:
            waiter.wait(
                StackName=stack_id, WaiterConfig={"Delay": 5, "MaxAttempts": 200}
            )
        except WaiterError as e:
            LOG.debug("Waiter failed for stack '%s'", stack_id, exc_info=e)
            LOG.critical(
                "Failed to create or update the '%s' stack. "
                "This stack is in your account, so you may be able to self-help by "
                "looking at '%s'. Otherwise, please reach out to CloudFormation.",
                stack_id,
                stack_id,
            )
            raise UploadError(
                "Failed to create or update the '{}' stack".format(stack_id)
            ) from e

        LOG.info(success_msg)

    def _get_stack_output(self, stack_id, output_key):
        result = self.cfn_client.describe_stacks(StackName=stack_id)
        outputs = result["Stacks"][0]["Outputs"]

        try:
            return next(
                output["OutputValue"]
                for output in outputs
                if output["OutputKey"] == output_key
            )
        except StopIteration:
            LOG.debug(
                "Outputs from stack '%s' did not contain '%s':\n%s",
                stack_id,
                output_key,
                ", ".join(output["OutputKey"] for output in outputs),
            )
            # pylint: disable=W0707
            raise InternalError("Required output not found on stack")

    def _create_or_update_stack(self, template, stack_name):
        args = {"StackName": stack_name, "TemplateBody": template}
        # attempt to create stack. if the stack already exists, try to update it
        LOG.info("Creating %s", stack_name)
        try:
            result = self.cfn_client.create_stack(
                **args,
                EnableTerminationProtection=True,
                Capabilities=["CAPABILITY_IAM"],
            )
        except self.cfn_client.exceptions.AlreadyExistsException:
            LOG.info("%s already exists. " "Attempting to update", stack_name)
            try:
                result = self.cfn_client.update_stack(
                    **args, Capabilities=["CAPABILITY_IAM"]
                )
            except ClientError as e:
                # if the update is a noop, don't do anything else
                msg = str(e)
                if "No updates are to be performed" in msg:
                    LOG.info("%s stack is up to date", stack_name)
                    stack_id = stack_name
                else:
                    LOG.debug(
                        "%s stack update " "resulted in unknown ClientError",
                        stack_name,
                        exc_info=e,
                    )
                    raise DownstreamError("Unknown CloudFormation error") from e
            else:
                stack_id = result["StackId"]
                self._wait_for_stack(
                    stack_id,
                    "stack_update_complete",
                    "{} stack is up to date".format(stack_name),
                )
        except ClientError as e:
            LOG.debug(
                "%s stack create " "resulted in unknown ClientError",
                stack_name,
                exc_info=e,
            )
            raise DownstreamError("Unknown CloudFormation error") from e
        else:
            stack_id = result["StackId"]
            self._wait_for_stack(
                stack_id,
                "stack_create_complete",
                "{} stack was successfully created".format(stack_name),
            )

        return stack_id

    def create_or_update_role(self, template_path, resource_type):
        try:
            with template_path.open("r", encoding="utf-8") as f:
                template = f.read()
        except FileNotFoundError:
            LOG.critical(
                "CloudFormation template 'resource-role.yaml' "
                "for execution role not found. "
                "Please run `generate` or "
                "provide an execution role via the --role-arn parameter."
            )
            # pylint: disable=W0707
            raise InvalidProjectError()
        stack_id = self._create_or_update_stack(
            template, "{}-role-stack".format(resource_type)
        )
        return self._get_stack_output(stack_id, EXECUTION_ROLE_ARN_OUTPUT_NAME)

    def upload(self, file_prefix, fileobj):
        template = self._get_template()
        stack_id = self._create_or_update_stack(template, INFRA_STACK_NAME)
        self.bucket_name = self._get_stack_output(stack_id, BUCKET_OUTPUT_NAME)
        self.log_delivery_role_arn = self._get_stack_output(
            stack_id, LOG_DELIVERY_ROLE_ARN_OUTPUT_NAME
        )

        timestamp = datetime.utcnow().isoformat(timespec="seconds").replace(":", "-")
        key = "{}-{}.zip".format(file_prefix, timestamp)

        LOG.debug("Uploading to '%s/%s'...", self.bucket_name, key)
        try:
            self.s3_client.upload_fileobj(fileobj, self.bucket_name, key)
        except ClientError as e:
            LOG.debug("S3 upload resulted in unknown ClientError", exc_info=e)
            raise DownstreamError("Failed to upload artifacts to S3") from e

        LOG.debug("Upload complete")

        return "s3://{0}/{1}".format(self.bucket_name, key)

    def get_log_delivery_role_arn(self):
        return self.log_delivery_role_arn



"""This sub command tests basic functionality of the resource in the project.

Projects can be created via the 'init' sub command.
"""
import json
import logging
import os
from argparse import SUPPRESS
from contextlib import contextmanager
from pathlib import Path
from tempfile import NamedTemporaryFile

import pytest
from jinja2 import Environment, Template, meta
from jsonschema import Draft6Validator
from jsonschema.exceptions import ValidationError

from rpdk.core.jsonutils.pointer import fragment_decode

from .boto_helpers import create_sdk_session, get_temporary_credentials
from .contract.contract_plugin import ContractPlugin
from .contract.interface import Action
from .contract.resource_client import ResourceClient
from .data_loaders import copy_resource
from .exceptions import SysExitRecommendedError
from .project import ARTIFACT_TYPE_MODULE, Project

LOG = logging.getLogger(__name__)

DEFAULT_ENDPOINT = "http://127.0.0.1:3001"
DEFAULT_FUNCTION = "TypeFunction"
DEFAULT_REGION = "us-east-1"
DEFAULT_TIMEOUT = "30"
INPUTS = "inputs"

OVERRIDES_VALIDATOR = Draft6Validator(
    {
        "properties": {"CREATE": {"type": "object"}, "UPDATE": {"type": "object"}},
        "anyOf": [{"required": ["CREATE"]}, {"required": ["UPDATE"]}],
        "additionalProperties": False,
    }
)


def empty_override():
    return {"CREATE": {}}


@contextmanager
def temporary_ini_file():
    with NamedTemporaryFile(
        mode="w", encoding="utf-8", prefix="pytest_", suffix=".ini"
    ) as temp:
        LOG.debug("temporary pytest.ini path: %s", temp.name)
        path = Path(temp.name).resolve(strict=True)
        copy_resource(__name__, "data/pytest-contract.ini", path)
        yield str(path)


def get_cloudformation_exports(region_name, endpoint_url, role_arn):
    session = create_sdk_session(region_name)
    temp_credentials = get_temporary_credentials(session, role_arn=role_arn)
    cfn_client = session.client(
        "cloudformation", endpoint_url=endpoint_url, **temp_credentials
    )
    paginator = cfn_client.get_paginator("list_exports")
    pages = paginator.paginate()
    exports = {}
    for page in pages:
        exports.update({export["Name"]: export["Value"] for export in page["Exports"]})
    return exports


def render_jinja(overrides_string, region_name, endpoint_url, role_arn):
    env = Environment(autoescape=True)
    parsed_content = env.parse(overrides_string)
    variables = meta.find_undeclared_variables(parsed_content)
    if variables:
        exports = get_cloudformation_exports(region_name, endpoint_url, role_arn)
        invalid_exports = variables - exports.keys()
        if len(invalid_exports) > 0:
            invalid_exports_message = (
                "Override file invalid: %s are not valid cloudformation exports."
                + "No Overrides will be applied"
            )
            LOG.warning(invalid_exports_message, invalid_exports)
            return empty_override()
        overrides_template = Template(overrides_string)
        to_return = json.loads(overrides_template.render(exports))
    else:
        to_return = json.loads(overrides_string)
    return to_return


def get_overrides(root, region_name, endpoint_url, role_arn):
    if not root:
        return empty_override()

    path = root / "overrides.json"
    try:
        with path.open("r", encoding="utf-8") as f:
            overrides_raw = render_jinja(f.read(), region_name, endpoint_url, role_arn)
    except FileNotFoundError:
        LOG.debug("Override file '%s' not found. No overrides will be applied", path)
        return empty_override()

    try:
        OVERRIDES_VALIDATOR.validate(overrides_raw)
    except ValidationError as e:
        LOG.warning("Override file invalid: %s\n" "No overrides will be applied", e)
        return empty_override()

    overrides = empty_override()
    for operation, items_raw in overrides_raw.items():
        items = {}
        for pointer, obj in items_raw.items():
            try:
                pointer = fragment_decode(pointer, prefix="")
            except ValueError:
                LOG.warning("%s pointer '%s' is invalid. Skipping", operation, pointer)
            else:
                items[pointer] = obj
        overrides[operation] = items

    return overrides


# pylint: disable=R0914
def get_inputs(root, region_name, endpoint_url, value, role_arn):
    inputs = {}
    if not root:
        return None

    path = root / INPUTS
    if not os.path.isdir(path):
        return None

    file_prefix = INPUTS + "_" + str(value)

    directories = os.listdir(path)
    if len(directories) > 0:
        for file in directories:
            if file.startswith(file_prefix) and file.endswith(".json"):
                input_type = get_type(file)
                if not input_type:
                    continue

                file_path = path / file
                with file_path.open("r", encoding="utf-8") as f:
                    overrides_raw = render_jinja(
                        f.read(), region_name, endpoint_url, role_arn
                    )
                overrides = {}
                for pointer, obj in overrides_raw.items():
                    overrides[pointer] = obj
                inputs[input_type] = overrides
        return inputs
    return None


def get_type(file_name):
    if "create" in file_name:
        return "CREATE"
    if "update" in file_name:
        return "UPDATE"
    if "invalid" in file_name:
        return "INVALID"
    return None


def get_marker_options(schema):
    lowercase_actions = {action.lower() for action in Action}
    excluded_actions = lowercase_actions - schema.get("handlers", {}).keys()
    marker_list = ["not " + action for action in excluded_actions]
    return " and ".join(marker_list)


def test(args):
    _validate_sam_args(args)
    project = Project()
    project.load()
    if project.artifact_type == ARTIFACT_TYPE_MODULE:
        LOG.warning("The test command is not supported in a module project")
        return

    overrides = get_overrides(
        project.root, args.region, args.cloudformation_endpoint_url, args.role_arn
    )

    index = 1
    while True:
        inputs = get_inputs(
            project.root,
            args.region,
            args.cloudformation_endpoint_url,
            index,
            args.role_arn,
        )
        if not inputs:
            break
        invoke_test(args, project, overrides, inputs)
        index = index + 1

    if index == 1:
        invoke_test(args, project, overrides, None)


def invoke_test(args, project, overrides, inputs):
    plugin = ContractPlugin(
        ResourceClient(
            args.function_name,
            args.endpoint,
            args.region,
            project.schema,
            overrides,
            inputs,
            args.role_arn,
            args.enforce_timeout,
            project.type_name,
            args.log_group_name,
            args.log_role_arn,
            executable_entrypoint=project.executable_entrypoint,
            docker_image=args.docker_image,
        )
    )

    with temporary_ini_file() as path:
        pytest_args = ["-c", path, "-m", get_marker_options(project.schema)]
        if args.passed_to_pytest:
            LOG.debug("extra args: %s", args.passed_to_pytest)
            pytest_args.extend(args.passed_to_pytest)
        LOG.debug("pytest args: %s", pytest_args)
        ret = pytest.main(pytest_args, plugins=[plugin])
        if ret:
            raise SysExitRecommendedError("One or more contract tests failed")


def setup_subparser(subparsers, parents):
    # see docstring of this file
    parser = subparsers.add_parser("test", description=__doc__, parents=parents)
    parser.set_defaults(command=test)

    _sam_arguments(parser)
    # this parameter can be used to pass additional arguments to pytest after `--`
    # for example,

    parser.add_argument(
        "--role-arn", help="Role used when performing handler operations."
    )

    parser.add_argument(
        "--cloudformation-endpoint-url", help="CloudFormation endpoint to use."
    )

    parser.add_argument(
        "--enforce-timeout",
        default=DEFAULT_TIMEOUT,
        help="Enforce a different timeout for handlers",
    )

    parser.add_argument(
        "--log-group-name",
        help="The log group to which contract tests lambda handler logs will be delivered. "
        "Specified log group doesn't have to exist as long as log-role-arn specified has logs:CreateLogGroup "
        "permission. Need to be used together with --log-role-arn",
    )

    parser.add_argument(
        "--log-role-arn",
        help="Role for delivering contract tests lambda handler logs. Need to be used together with --log-group-name",
    )

    parser.add_argument("passed_to_pytest", nargs="*", help=SUPPRESS)

    parser.add_argument(
        "--docker-image",
        help="Docker image name to run. If specified, invoke will use docker instead "
        "of SAM",
    )


def _sam_arguments(parser):
    parser.add_argument(
        "--endpoint",
        default=DEFAULT_ENDPOINT,
        help=(
            "The endpoint at which the type can be invoked "
            f"(Default: {DEFAULT_ENDPOINT})"
        ),
    )
    parser.add_argument(
        "--function-name",
        default=DEFAULT_FUNCTION,
        help=(
            "The logical lambda function name in the SAM template "
            f"(Default: {DEFAULT_FUNCTION})"
        ),
    )
    parser.add_argument(
        "--region",
        default=DEFAULT_REGION,
        help=(
            "The region used for temporary credentials " f"(Default: {DEFAULT_REGION})"
        ),
    )


def _validate_sam_args(args):
    if args.docker_image and (
        args.endpoint != DEFAULT_ENDPOINT or args.function_name != DEFAULT_FUNCTION
    ):
        raise SysExitRecommendedError(
            "Cannot specify both --docker-image and --endpoint or --function-name"
        )



#!/usr/bin/env python
import os.path
import re

from setuptools import setup

HERE = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    with open(os.path.join(HERE, *parts), "r", encoding="utf-8") as fp:
        return fp.read()


# https://packaging.python.org/guides/single-sourcing-package-version/
def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


setup(
    name="cloudformation-cli",
    version=find_version("src", "rpdk", "core", "__init__.py"),
    description=__doc__,
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    author="Amazon Web Services",
    author_email="aws-cloudformation-developers@amazon.com",
    url="https://github.com/aws-cloudformation/aws-cloudformation-rpdk/",
    # https://packaging.python.org/guides/packaging-namespace-packages/
    packages=["rpdk.core"],
    package_dir={"": "src"},
    # package_data -> use MANIFEST.in instead
    include_package_data=True,
    zip_safe=True,
    python_requires=">=3.6",
    install_requires=[
        "boto3>=1.10.20",
        "Jinja2>=2.10,<3",
        "jsonschema>=3.0.1,<4.0",
        "pytest>=4.5.0",
        "pytest-random-order>=1.0.4",
        "pytest-localserver>=0.5.0",
        "Werkzeug>=0.15",
        "PyYAML>=5.1",
        "requests>=2.22",
        "hypothesis>=4.32",
        "colorama>=0.4.1",
        "docker>=4.3.1",
        "ordered-set>=4.0.2",
        "cfn-lint>=0.43.0",
        "cfn_flip>=1.2.3",
        "nested-lookup",
    ],
    entry_points={
        "console_scripts": ["cfn-cli = rpdk.core.cli:main", "cfn = rpdk.core.cli:main"]
    },
    license="Apache License 2.0",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Topic :: Software Development :: Build Tools",
        "Topic :: Software Development :: Code Generators",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    keywords="Amazon Web Services AWS CloudFormation",
)



"""This sub command adds credentials to a payload, and sends it to a Lambda
function. The function is re-invoked while the IN_PROGRESS status is returned.

Projects can be created via the 'init' sub command.
"""
# pylint: disable=protected-access
import json
import logging
from argparse import FileType
from time import sleep

from .contract.interface import Action, OperationStatus
from .contract.resource_client import ResourceClient
from .exceptions import SysExitRecommendedError
from .project import Project
from .test import _sam_arguments, _validate_sam_args

LOG = logging.getLogger(__name__)


def invoke(args):
    _validate_sam_args(args)
    project = Project()
    project.load()

    client = ResourceClient(
        args.function_name,
        args.endpoint,
        args.region,
        project.schema,
        {},
        executable_entrypoint=project.executable_entrypoint,
        docker_image=args.docker_image,
    )

    action = Action[args.action]
    try:
        request = json.load(args.request)
    except ValueError as e:
        raise SysExitRecommendedError(f"Invalid JSON: {e}") from e
    payload = client._make_payload(
        action,
        request["desiredResourceState"],
        request["previousResourceState"],
        request.get("typeConfiguration"),
    )

    current_invocation = 0

    try:
        while _needs_reinvocation(args.max_reinvoke, current_invocation):
            print("=== Handler input ===")

            payload_to_log = {
                "callbackContext": payload["callbackContext"],
                "action": payload["action"],
                "requestData": {
                    "resourceProperties": payload["requestData"]["resourceProperties"],
                    "previousResourceProperties": payload["requestData"][
                        "previousResourceProperties"
                    ],
                },
                "region": payload["region"],
                "awsAccountId": payload["awsAccountId"],
                "bearerToken": payload["bearerToken"],
            }
            print(json.dumps({**payload_to_log}, indent=2))

            response = client._call(payload)
            current_invocation = current_invocation + 1
            print("=== Handler response ===")
            print(json.dumps(response, indent=2))
            status = OperationStatus[response["status"]]

            if status != OperationStatus.IN_PROGRESS:
                break

            sleep(response.get("callbackDelaySeconds", 0))
            payload["callbackContext"] = response.get("callbackContext")
    except KeyboardInterrupt:
        pass


def _needs_reinvocation(max_reinvoke, current_invocation):
    return max_reinvoke is None or max_reinvoke >= current_invocation


def setup_subparser(subparsers, parents):
    # see docstring of this file
    parser = subparsers.add_parser("invoke", description=__doc__, parents=parents)
    parser.set_defaults(command=invoke)
    parser.add_argument(
        "action",
        choices=list(Action.__members__),
        help="The provisioning action, i.e. which handler to invoke.",
    )
    parser.add_argument(
        "request",
        type=FileType("r", encoding="utf-8"),
        help="A JSON file that contains the request to invoke the function with.",
    )
    parser.add_argument(
        "--max-reinvoke",
        type=int,
        default=None,
        help="Maximum number of IN_PROGRESS re-invocations allowed before "
        "exiting. If not specified, will continue to "
        "re-invoke until terminal status is reached.",
    )
    parser.add_argument(
        "--docker-image",
        help="Docker image name to run. If specified, invoke will use docker instead "
        "of SAM",
    )
    _sam_arguments(parser)



import copy
import logging

from rpdk.core.contract.interface import Action, HandlerErrorCode, OperationStatus
from rpdk.core.contract.resource_client import (
    prune_properties_from_model,
    prune_properties_if_not_exist_in_path,
)
from rpdk.core.contract.suite.contract_asserts import (
    failed_event,
    response_contains_primary_identifier,
    response_contains_resource_model_equal_updated_model,
    response_contains_unchanged_primary_identifier,
    response_does_not_contain_write_only_properties,
)

LOG = logging.getLogger(__name__)


@response_contains_primary_identifier
@response_does_not_contain_write_only_properties
def test_create_success(resource_client, current_resource_model):
    _status, response, _error_code = resource_client.call_and_assert(
        Action.CREATE, OperationStatus.SUCCESS, current_resource_model
    )
    return response


@failed_event(
    error_code=HandlerErrorCode.AlreadyExists,
    msg="A create handler MUST NOT create multiple resources given\
         the same idempotency token",
)
def test_create_failure_if_repeat_writeable_id(resource_client, current_resource_model):
    LOG.debug(
        "at least one identifier is writeable; "
        "performing duplicate-CREATE-failed test"
    )
    # Should fail, because different clientRequestToken for the same
    # resource model means that the same resource is trying to be
    # created twice.
    _status, _response, error_code = resource_client.call_and_assert(
        Action.CREATE, OperationStatus.FAILED, current_resource_model
    )
    return error_code


@response_contains_primary_identifier
@response_does_not_contain_write_only_properties
def test_read_success(resource_client, current_resource_model):
    _status, response, _error_code = resource_client.call_and_assert(
        Action.READ, OperationStatus.SUCCESS, current_resource_model
    )
    test_input_equals_output(
        resource_client, current_resource_model, response["resourceModel"]
    )
    return response


@failed_event(
    error_code=HandlerErrorCode.NotFound,
    msg="A read handler MUST return FAILED with a NotFound error code\
         if the resource does not exist",
)
def test_read_failure_not_found(
    resource_client,
    current_resource_model,
):
    _status, _response, error_code = resource_client.call_and_assert(
        Action.READ, OperationStatus.FAILED, current_resource_model
    )
    return error_code


def get_resource_model_list(resource_client, current_resource_model):
    _status, response, _error_code = resource_client.call_and_assert(
        Action.LIST, OperationStatus.SUCCESS, current_resource_model
    )
    next_token = response.get("nextToken")
    resource_models = response["resourceModels"]
    while next_token is not None:
        _status, next_response, _error_code = resource_client.call_and_assert(
            Action.LIST,
            OperationStatus.SUCCESS,
            current_resource_model,
            nextToken=next_token,
        )
        resource_models.extend(next_response["resourceModels"])
        next_token = next_response.get("nextToken")
    return resource_models


def test_model_in_list(resource_client, current_resource_model):
    resource_models = get_resource_model_list(resource_client, current_resource_model)
    return any(
        resource_client.is_primary_identifier_equal(
            resource_client.primary_identifier_paths,
            resource_model,
            current_resource_model,
        )
        for resource_model in resource_models
    )


@response_contains_primary_identifier
@response_contains_unchanged_primary_identifier
@response_contains_resource_model_equal_updated_model
@response_does_not_contain_write_only_properties
def test_update_success(resource_client, update_resource_model, current_resource_model):
    _status, response, _error_code = resource_client.call_and_assert(
        Action.UPDATE,
        OperationStatus.SUCCESS,
        update_resource_model,
        current_resource_model,
    )
    return response


@failed_event(
    error_code=HandlerErrorCode.NotFound,
    msg="An update handler MUST return FAILED with a NotFound error code\
         if the resource did not exist prior to the update request",
)
def test_update_failure_not_found(resource_client, current_resource_model):
    update_model = resource_client.generate_update_example(current_resource_model)
    _status, _response, error_code = resource_client.call_and_assert(
        Action.UPDATE, OperationStatus.FAILED, update_model, current_resource_model
    )
    return error_code


def test_delete_success(resource_client, current_resource_model):
    _status, response, _error_code = resource_client.call_and_assert(
        Action.DELETE, OperationStatus.SUCCESS, current_resource_model
    )
    return response


@failed_event(
    error_code=HandlerErrorCode.NotFound,
    msg="A delete hander MUST return FAILED with a NotFound error code\
         if the resource did not exist prior to the delete request",
)
def test_delete_failure_not_found(resource_client, current_resource_model):
    _status, _response, error_code = resource_client.call_and_assert(
        Action.DELETE, OperationStatus.FAILED, current_resource_model
    )
    return error_code


def test_input_equals_output(resource_client, input_model, output_model):
    pruned_input_model = prune_properties_from_model(
        copy.deepcopy(input_model),
        set(
            list(resource_client.read_only_paths)
            + list(resource_client.write_only_paths)
        ),
    )

    pruned_output_model = prune_properties_from_model(
        copy.deepcopy(output_model), resource_client.read_only_paths
    )

    pruned_output_model = prune_properties_if_not_exist_in_path(
        pruned_output_model, pruned_input_model, resource_client.create_only_paths
    )
    resource_client.compare(
        pruned_input_model,
        pruned_output_model,
    )



"""Encode JavaScript Object Notation (JSON) Pointer as per
`RFC-6901 <https://tools.ietf.org/html/rfc6901>`_.
"""
from itertools import chain
from urllib.parse import quote, unquote


def part_encode(part):
    """Encode a part of a JSON pointer.

    >>> part_encode("foo")
    'foo'
    >>> part_encode("~foo")
    '~0foo'
    >>> part_encode("foo~")
    'foo~0'
    >>> part_encode("/foo")
    '~1foo'
    >>> part_encode("foo/")
    'foo~1'
    >>> part_encode("f/o~o")
    'f~1o~0o'
    >>> part_encode("~0")
    '~00'
    >>> part_encode("~1")
    '~01'
    >>> part_encode(0)
    '0'
    """
    return str(part).replace("~", "~0").replace("/", "~1")


def part_decode(part):
    """Decode a part of a JSON pointer.

    >>> part_decode("foo")
    'foo'
    >>> part_decode("~0foo")
    '~foo'
    >>> part_decode("foo~0")
    'foo~'
    >>> part_decode("~1foo")
    '/foo'
    >>> part_decode("foo~1")
    'foo/'
    >>> part_decode("f~1o~0o")
    'f/o~o'
    >>> part_decode("~00")
    '~0'
    >>> part_decode("~01")
    '~1'
    >>> part_decode("0")
    '0'
    """
    return part.replace("~1", "/").replace("~0", "~")


def fragment_encode(parts, prefix="#"):
    """Encode all parts of a JSON pointer into the URI fragment
    identifier representation.

    >>> fragment_encode([])
    '#'
    >>> fragment_encode([], prefix="")
    ''
    >>> fragment_encode(["foo", "bar"])
    '#/foo/bar'
    >>> fragment_encode([0, " ", "~"])
    '#/0/%20/~0'
    """
    encoded = (quote(part_encode(part), safe="/~") for part in parts)
    return "/".join(chain([prefix], encoded))


def fragment_decode(pointer, prefix="#", output=tuple):
    """Decode all segments of a JSON pointer from the URI fragment
    identifier representation.

    >>> fragment_decode("#")
    ()
    >>> fragment_decode("#/foo/bar")
    ('foo', 'bar')
    >>> fragment_decode("#/foo/bar", output=list)
    ['foo', 'bar']
    >>> fragment_decode("#/0/%20/~0")
    ('0', ' ', '~')
    >>> fragment_decode("/foo")
    Traceback (most recent call last):
    ...
    ValueError: Expected prefix '#', but was ''
    """
    segments = pointer.split("/")
    decoded = (part_decode(unquote(segment)) for segment in segments)
    actual = next(decoded)
    if prefix != actual:
        raise ValueError("Expected prefix '{}', but was '{}'".format(prefix, actual))
    return output(decoded)


def fragment_list(segments, prefix="properties", output=list):
    """Decode all segments of a JSON pointer from the URI fragment
    identifier representation.

    >>> fragment_list(["properties"])
    []
    >>> fragment_list(["properties", "foo", "bar"])
    ['foo', 'bar']
    >>> fragment_list(["properties", "foo", "bar"], output=tuple)
    ('foo', 'bar')
    >>> fragment_list(["properties", "0", "%20", "~0"])
    ['0', ' ', '~']
    >>> fragment_list(["foo"])
    Traceback (most recent call last):
    ...
    ValueError: Expected prefix 'properties', but was 'foo'
    """
    decoded = (part_decode(unquote(segment)) for segment in segments)
    actual = next(decoded)
    if prefix != actual:
        raise ValueError("Expected prefix '{}', but was '{}'".format(prefix, actual))
    return output(decoded)



import json
import logging

import cfnlint.config
import cfnlint.core

from .module_fragment_reader import _get_fragment_file

LOG = logging.getLogger(__name__)


def print_cfn_lint_warnings(fragment_dir):
    lint_warnings = __get_cfn_lint_matches(fragment_dir)
    if not lint_warnings:
        LOG.warning("Module fragment is valid.")
    else:
        LOG.warning(
            "Module fragment might be valid, but there are "
            "warnings from cfn-lint "
            "(https://github.com/aws-cloudformation/cfn-python-lint):"
        )
        for lint_warning in lint_warnings:
            print(
                "\t{} (from rule {})".format(lint_warning.message, lint_warning.rule),
            )


def __get_cfn_lint_matches(fragment_dir):
    filepath = _get_fragment_file(fragment_dir)
    try:
        try:
            template = cfnlint.decode.cfn_json.load(filepath)
        except json.decoder.JSONDecodeError:
            template = cfnlint.decode.cfn_yaml.load(filepath)
    except Exception as e:  # pylint: disable=broad-except
        LOG.error(
            "Skipping cfn-lint validation due to an internal error.\n"
            "Please report this issue to the team (include rpdk.log file)\n"
            "Issue tracker: github.com/aws-cloudformation/cloudformation-cli/issues"
        )
        LOG.error(str(e))
        return []

    # Initialize the ruleset to be applied (no overrules, no excludes)
    rules = cfnlint.core.get_rules([], [], [], [], False, [])

    # Default region used by cfn-lint
    regions = ["us-east-1"]

    # Runs Warning and Error rules
    matches = cfnlint.core.run_checks(filepath, template, rules, regions)

    return matches
